#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.device.framebuffer` class.
"""
import os
import pytest

from luma.core.render import canvas
from luma.core.framebuffer import full_frame
from luma.core.device import linux_framebuffer
import luma.core.error

from helpers import multi_mock_open, get_reference_file
from unittest.mock import patch, call

WIDTH = 124
HEIGHT = 55
SCREEN_RES = f"{WIDTH},{HEIGHT}"
BITS_PER_PIXEL = "24"


def swap_red_and_blue(data, step):
    arr = bytearray(data)
    for i in range(0, len(arr), step):
        [red, green, blue] = arr[i: i + 3]
        arr[i: i + 3] = [blue, green, red]
    return bytes(arr)


def test_display_id_as_dev_fb_number():
    with patch("builtins.open", multi_mock_open(SCREEN_RES, BITS_PER_PIXEL, None)):
        device = linux_framebuffer("/dev/fb9")
        assert device.id == 9


def test_display_id_from_environ():
    os.environ["FRAMEBUFFER"] = "/dev/fb16"
    with patch("builtins.open", multi_mock_open(SCREEN_RES, BITS_PER_PIXEL, None)):
        device = linux_framebuffer()
        assert device.id == 16


def test_unknown_display_id():
    with patch("builtins.open", multi_mock_open(SCREEN_RES, BITS_PER_PIXEL, None)):
        with pytest.raises(luma.core.error.DeviceNotFoundError):
            linux_framebuffer("invalid fb")


def test_read_screen_resolution():
    with patch(
        "builtins.open", multi_mock_open(SCREEN_RES, BITS_PER_PIXEL, None)
    ) as fake_open:
        device = linux_framebuffer("/dev/fb1")
        assert device.width == 124
        assert device.height == 55
        fake_open.assert_has_calls([call("/sys/class/graphics/fb1/virtual_size", "r")])


def test_read_bits_per_pixel():
    with patch(
        "builtins.open", multi_mock_open(SCREEN_RES, BITS_PER_PIXEL, None)
    ) as fake_open:
        device = linux_framebuffer("/dev/fb1")
        assert device.bits_per_pixel == 24
        fake_open.assert_has_calls(
            [call("/sys/class/graphics/fb1/bits_per_pixel", "r")]
        )


@pytest.mark.parametrize("bits_per_pixel,bgr", [
    (16, False),
    (24, False),
    (24, True),
    (32, False),
    (32, True),
])
def test_display(bits_per_pixel, bgr):
    bytes_per_pixel = bits_per_pixel // 8
    with open(get_reference_file(f"fb_{bits_per_pixel}bpp.raw"), "rb") as fp:
        reference = fp.read()
        if bgr:
            reference = swap_red_and_blue(reference, step=bytes_per_pixel)

    with patch("builtins.open", multi_mock_open(SCREEN_RES, str(bits_per_pixel), None)) as fake_open:
        device = linux_framebuffer("/dev/fb1", framebuffer=full_frame(), bgr=bgr)

        fake_open.assert_has_calls([call("/dev/fb1", "wb")])
        fake_open.reset_mock()

        with canvas(device, dither=True) as draw:
            draw.rectangle((0, 0, 64, 32), fill="red")
            draw.rectangle((64, 0, 128, 32), fill="yellow")
            draw.rectangle((0, 32, 64, 64), fill="orange")
            draw.rectangle((64, 32, 128, 64), fill="white")

        fake_open.return_value.seek.assert_has_calls([
            call(n * WIDTH * bytes_per_pixel)
            for n in range(HEIGHT)
        ])
        fake_open.return_value.write.assert_has_calls([
            call(reference[n:n + (WIDTH * bytes_per_pixel)])
            for n in range(0, len(reference), WIDTH * bytes_per_pixel)
        ])
        fake_open.return_value.flush.assert_called_once()


def test_unsupported_bit_depth():
    with patch("builtins.open", multi_mock_open(SCREEN_RES, "19", None)):
        with pytest.raises(AssertionError) as ex:
            linux_framebuffer("/dev/fb4")
        assert str(ex.value) == 'Unsupported bit-depth: 19'


def test_cleanup():
    with patch("builtins.open", multi_mock_open(SCREEN_RES, BITS_PER_PIXEL, None)) as fake_open:
        device = linux_framebuffer("/dev/fb1", framebuffer=full_frame())
        device.cleanup()
        fake_open.return_value.close.assert_called_once()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.interface.serial.pcf8574` class.
"""

import errno
import pytest

from unittest.mock import Mock, call
from luma.core.interface.serial import pcf8574
import luma.core.error


smbus = Mock(unsafe=True)
BACKLIGHT = 0x08
ENABLE = 0x04
COMMAND = 0x00
DATA = 0x01


def setup_function(function):
    smbus.reset_mock()


def test_command():
    cmds = [3, 1, 4]
    serial = pcf8574(bus=smbus, address=0x27)
    serial.command(*cmds)

    calls = []
    for c in cmds:
        calls += [call(0x27, BACKLIGHT | COMMAND | c << 4)]
        calls += [call(0x27, BACKLIGHT | COMMAND | c << 4 | ENABLE)]
        calls += [call(0x27, BACKLIGHT | COMMAND | c << 4)]

    smbus.write_byte.assert_has_calls(calls)


def test_i2c_command_device_not_found_error():
    errorbus = Mock(unsafe=True)
    address = 0x27
    cmds = [3, 1, 4]
    expected_error = OSError()

    try:
        for error_code in [errno.EREMOTEIO, errno.EIO]:
            expected_error.errno = error_code
            errorbus.write_byte.side_effect = expected_error

            serial = pcf8574(bus=errorbus, address=address)
            with pytest.raises(luma.core.error.DeviceNotFoundError) as ex:
                serial.command(*cmds)

            assert str(ex.value) == f'I2C device not found on address: 0x{address:02X}'
    except AttributeError as e:
        # osx
        pytest.skip(str(e))


def test_i2c_data():
    data = list((5, 4, 3, 2, 1))
    serial = pcf8574(bus=smbus, address=0x21)
    serial.data(data)

    calls = []
    for d in data:
        calls += [call(0x21, BACKLIGHT | DATA | d << 4)]
        calls += [call(0x21, BACKLIGHT | DATA | d << 4 | ENABLE)]
        calls += [call(0x21, BACKLIGHT | DATA | d << 4)]

    smbus.write_byte.assert_has_calls(calls)


def test_cleanup():
    serial = pcf8574(bus=smbus, address=0x9F)
    serial._managed = True
    serial.cleanup()
    smbus.close.assert_called_once_with()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:module:`luma.core.interface.parallel` module.
"""

from unittest.mock import Mock, call
from luma.core.interface.parallel import bitbang_6800
import luma.core.error

import pytest

from helpers import rpi_gpio_missing, assert_only_cleans_whats_setup


gpio = Mock(unsafe=True)


def setup_function(function):
    gpio.reset_mock()
    gpio.HIGH = 200
    gpio.LOW = 100
    gpio.RS = 7
    gpio.E = 8
    gpio.PINS = [25, 24, 23, 18]
    gpio.DATA = gpio.HIGH
    gpio.CMD = gpio.LOW
    gpio.OUT = 300


def test_data():
    eight_to_four = lambda data: [f(x) for x in data for f in (lambda x: x >> 4, lambda x: 0x0F & x)]

    data = (0x41, 0x42, 0x43)  # ABC
    serial = bitbang_6800(gpio=gpio, RS=7, E=8, PINS=[25, 24, 23, 18])

    serial.command(*eight_to_four([0x80]))
    serial.data(eight_to_four(data))

    setup = [call(gpio.RS, gpio.OUT), call(gpio.E, gpio.OUT)] + \
        [call(gpio.PINS[i], gpio.OUT) for i in range(4)]
    prewrite = lambda mode: [call(gpio.RS, mode), call(gpio.E, gpio.LOW)]
    pulse = [call(gpio.E, gpio.HIGH), call(gpio.E, gpio.LOW)]
    send = lambda v: [call(gpio.PINS[i], (v >> i) & 0x01) for i in range(serial._datalines)]

    calls = \
        prewrite(gpio.CMD) + send(0x08) + pulse + send(0x00) + pulse + \
        prewrite(gpio.DATA) + \
        send(data[0] >> 4) + pulse + \
        send(data[0]) + pulse + \
        send(data[1] >> 4) + pulse + \
        send(data[1]) + pulse + \
        send(data[2] >> 4) + pulse + \
        send(data[2]) + pulse

    gpio.setup.assert_has_calls(setup)
    gpio.output.assert_has_calls(calls)


def test_wrong_number_of_pins():
    try:
        bitbang_6800(gpio=gpio, RS=7, E=8, PINS=[25, 24, 23])
    except AssertionError as ex:
        assert str(ex) == 'You\'ve provided 3 pins but a bus must contain either four or eight pins'


def test_cleanup():
    serial = bitbang_6800(gpio=gpio)
    serial._managed = True
    serial.cleanup()
    assert_only_cleans_whats_setup(gpio)


def test_unsupported_gpio_platform():
    try:
        bitbang_6800()
    except luma.core.error.UnsupportedPlatform as ex:
        assert str(ex) == 'GPIO access not available'
    except ImportError:
        pytest.skip(rpi_gpio_missing)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.


from luma.core.legacy.font import proportional, tolerant, CP437_FONT


def test_default():
    font = tolerant(CP437_FONT)
    assert font[65] == [0x7C, 0x7E, 0x13, 0x13, 0x7E, 0x7C, 0x00, 0x00]
    assert font[6543] == [0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80]


def test_custom_missing():
    font = tolerant(CP437_FONT, missing="?")
    assert font[65] == [0x7C, 0x7E, 0x13, 0x13, 0x7E, 0x7C, 0x00, 0x00]
    assert font[6543] == [0x02, 0x03, 0x51, 0x59, 0x0F, 0x06, 0x00, 0x00]


def test_with_proportional():
    font = proportional(tolerant(CP437_FONT, missing="?"))
    assert font[65] == [0x7C, 0x7E, 0x13, 0x13, 0x7E, 0x7C, 0x00]
    assert font[6543] == [0x02, 0x03, 0x51, 0x59, 0x0F, 0x06, 0x00]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.interface.serial.bitbang` class.
"""

from unittest.mock import Mock, call
from luma.core.interface.serial import bitbang
import luma.core.error

import pytest

from helpers import rpi_gpio_missing, assert_only_cleans_whats_setup


gpio = Mock(unsafe=True)


def setup_function(function):
    gpio.reset_mock()
    gpio.BCM = 1
    gpio.RST = 2
    gpio.DC = 3
    gpio.OUT = 4
    gpio.HIGH = 5
    gpio.LOW = 6


def test_data():
    data = (0xFF, 0x0F, 0x00)
    serial = bitbang(gpio=gpio, SCLK=13, SDA=14, CE=15, DC=16, RST=17)
    serial.data(data)

    reset = [call(17, gpio.LOW), call(17, gpio.HIGH)]
    clock = [call(13, gpio.HIGH), call(13, gpio.LOW)]
    data = lambda x: call(14, 0x80 if x == gpio.HIGH else 0x00)
    ce = lambda x: [call(15, x)]
    dc = lambda x: [call(16, x)]

    calls = reset + \
        dc(gpio.HIGH) + \
        ce(gpio.LOW) + \
        (([data(gpio.HIGH)] + clock) * 8) + \
        (([data(gpio.LOW)] + clock) * 4) + \
        (([data(gpio.HIGH)] + clock) * 4) + \
        (([data(gpio.LOW)] + clock) * 8) + \
        ce(gpio.HIGH)

    gpio.output.assert_has_calls(calls)


def test_cleanup():
    serial = bitbang(gpio=gpio)
    serial._managed = True
    serial.cleanup()
    assert_only_cleans_whats_setup(gpio)


def test_cleanup_custom_pins():
    serial = bitbang(gpio=gpio, SCLK=13, SDA=14, CE=15, DC=16, RST=17)
    serial._managed = True
    serial.cleanup()
    assert_only_cleans_whats_setup(gpio)


def test_unsupported_gpio_platform():
    try:
        bitbang()
    except luma.core.error.UnsupportedPlatform as ex:
        assert str(ex) == 'GPIO access not available'
    except ImportError:
        pytest.skip(rpi_gpio_missing)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.


from luma.core.util import observable, mutable_string


class test_bell(object):

    called = 0

    def ding(self, *args):
        self.called += 1


def test_length():
    bell = test_bell()
    buf = observable(bytearray("hello", "utf-8"), bell.ding)
    assert len(buf) == 5
    assert bell.called == 1


def test_iteration():
    bell = test_bell()
    buf = observable(bytearray("hello", "utf-8"), bell.ding)
    assert list(iter(buf)) == [ord(ch) for ch in "hello"]
    assert bell.called == 1


def test_getattribute():
    bell = test_bell()
    buf = observable(bytearray("hello", "utf-8"), bell.ding)
    assert list(iter(buf.decode("utf-8"))) == list("hello")
    assert bell.called == 1


def test_getitem():
    bell = test_bell()
    buf = observable(mutable_string("hello"), bell.ding)
    assert buf[2] == "l"
    assert bell.called == 1


def test_setitem():
    bell = test_bell()
    buf = observable(mutable_string("hello"), bell.ding)
    buf[0] = "y"
    assert str(buf) == "yello"
    assert bell.called == 2


def test_setslice():
    bell = test_bell()
    buf = observable(mutable_string("hello"), bell.ding)
    buf[1:4] = "ipp"
    assert str(buf) == "hippo"
    assert bell.called == 2


def test_delitem():
    bell = test_bell()
    buf = observable(mutable_string("hello"), bell.ding)
    del buf[4]
    assert str(buf) == "hell"
    assert bell.called == 2


def test_getslice():
    bell = test_bell()
    buf = observable(mutable_string("hello"), bell.ding)
    assert buf[2:4] == "ll"
    assert bell.called == 1


def test_repr():
    bell = test_bell()
    buf = observable(bytearray("hello", "utf-8"), bell.ding)
    assert repr(buf) == "bytearray(b'hello')"
    assert bell.called == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.image_composition` class and associated
helpers.
"""

import pytest
import PIL

from PIL import Image, ImageDraw
from luma.core.device import dummy
from luma.core.image_composition import ComposableImage, ImageComposition


def test_composable_image_none():
    with pytest.raises(AssertionError):
        ComposableImage(None)


def test_composable_image_image():
    ci = ComposableImage(Image.new("RGB", (1, 1)))
    with pytest.raises(AssertionError):
        ci.image((0, 0))

    with pytest.raises(AssertionError):
        ci.image((1, 0))

    with pytest.raises(AssertionError):
        ci.image((0, 1))


def test_composable_image_ctor():
    pos = (78, 12)
    offs = (90, 12)
    img_size = (123, 234)
    img = Image.new("RGB", img_size)
    ci = ComposableImage(img)
    ci.position = pos
    ci.offset = offs
    assert ci.position == pos
    assert ci.offset == offs
    assert ci.width == img_size[0]
    assert ci.height == img_size[1]
    assert isinstance(ci.image((1, 1)), PIL.Image.Image)


def test_composable_image_crop_same():
    img_size = (128, 64)
    crop_size = img_size
    img = Image.new("RGB", img_size)
    ci = ComposableImage(img)
    cropped_img = ci.image(crop_size)
    assert cropped_img.size == crop_size


def test_composable_image_crop_size_smaller_than_image_size():
    img_size = (128, 64)
    crop_size = (10, 10)
    img = Image.new("RGB", img_size)
    ci = ComposableImage(img)
    cropped_img = ci.image(crop_size)
    assert cropped_img.size == crop_size


def test_composable_image_crop_size_greater_than_image_size():
    img_size = (128, 64)
    crop_size = (1000, 1000)
    img = Image.new("RGB", img_size)
    ci = ComposableImage(img)
    cropped_img = ci.image(crop_size)
    assert cropped_img.size == img_size


def test_composable_image_crop_offset():
    img_size = (128, 64)
    crop_size = (10, 10)
    img = Image.new("RGB", img_size)
    ci = ComposableImage(img, offset=(20, 20))
    cropped_img = ci.image(crop_size)
    assert cropped_img.size == crop_size


def test_image_composition_ctor():
    ic = ImageComposition(dummy())
    assert isinstance(ic(), PIL.Image.Image)


def test_image_add_image_none():
    ic = ImageComposition(dummy())
    with pytest.raises(AssertionError):
        ic.add_image(None)


def test_image_remove_image_none():
    ic = ImageComposition(dummy())
    with pytest.raises(AssertionError):
        ic.remove_image(None)


def test_image_count():
    ic = ImageComposition(dummy())
    img1 = ComposableImage(Image.new("RGB", (1, 1)))
    img2 = ComposableImage(Image.new("RGB", (1, 1)))
    img3 = ComposableImage(Image.new("RGB", (1, 1)))
    ic.add_image(img1)
    ic.add_image(img2)
    ic.add_image(img3)
    assert len(ic.composed_images) == 3
    ic.remove_image(img3)
    assert len(ic.composed_images) == 2
    ic.remove_image(img2)
    assert len(ic.composed_images) == 1
    ic.remove_image(img1)
    assert len(ic.composed_images) == 0


def test_refresh_no_images():
    ic = ImageComposition(dummy())
    ic_img_before = list(ic().getdata())
    ic.refresh()
    assert ic_img_before == list(ic().getdata())


def test_refresh():
    ic = ImageComposition(dummy())
    ic_img_before = list(ic().getdata())
    img = Image.new("RGB", (25, 25))
    draw = ImageDraw.Draw(img)
    draw.rectangle((10, 10, 20, 20), outline="white")
    del draw
    ci = ComposableImage(img)
    ic.add_image(ci)
    ic.refresh()
    assert ic_img_before != list(ic().getdata())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.virtual.terminal` class.
"""

from PIL import Image

from luma.core.device import dummy
from luma.core.virtual import terminal

from helpers import (get_reference_image, assert_identical_image,
    get_reference_font, test_font)


def assert_text(device, term, reference_img, text, save=None):
    img_path = get_reference_image(reference_img)

    with open(img_path, 'rb') as fp:
        reference = Image.open(fp)

        for line in text:
            term.println(line)

        if save is not None:
            device.image.save(save)

        assert_identical_image(reference, device.image, img_path)


def test_default_text():
    reference = 'quick_brown_fox.png'
    device = dummy()
    term = terminal(device, font=test_font)

    assert_text(device, term, reference, [
        "The quick brown fox jumps over the lazy dog"
    ])


def test_wrapped_text():
    reference = 'quick_brown_fox_word_wrap.png'
    device = dummy()
    term = terminal(device, font=test_font, word_wrap=True, animate=False)

    assert_text(device, term, reference, [
        "The quick brown fox jumps over the lazy dog"
    ])


def test_tab_alignment():
    reference = 'tab_align.png'
    device = dummy()
    term = terminal(device, font=test_font, animate=False)

    assert_text(device, term, reference, [
        "1\t32\t999",
        "999\t1\t32"
    ])


def test_control_chars():
    reference = 'control_chars.png'
    device = dummy()
    term = terminal(device, font=test_font, animate=False)

    assert_text(device, term, reference, [
        'foo\rbar\bspam\teggs\n\nham and cheese on rye'
    ])


def test_scrolling():
    reference = 'scroll_text.png'
    device = dummy()
    term = terminal(device, font=test_font, animate=False)

    assert_text(device, term, reference, [
        "it oozed over the blackness, and heard Harris's sleepy voice asking "
        "where we drew near it, so they spread their handkerchiefs on the back "
        "of Harris and Harris's friend as to avoid running down which, we managed "
        "to get out of here while this billing and cooing is on. We'll go down "
        "to eat vegetables. He said they were demons."
    ])


def test_alt_colors():
    reference = 'alt_colors.png'
    device = dummy()
    term = terminal(device, font=test_font, color="blue", bgcolor="grey", animate=False)

    assert_text(device, term, reference, [
        "blue on grey"
    ])


def test_ansi_colors():
    reference = 'ansi_colors.png'
    device = dummy()
    term = terminal(device, font=test_font, animate=False)

    assert_text(device, term, reference, [
        "hello \033[31mworld\033[0m ansi colors here!",
        "this is \033[7mreversed\033[7m!",
        "\033[44;37mBlue\033[0m \033[46;30mCyan"
    ])


def test_ansi_colors_wrapped():
    reference = 'ansi_colors_wrapped.png'
    device = dummy()
    term = terminal(device, font=test_font, word_wrap=True, animate=False)

    assert_text(device, term, reference, [
        "hello \033[31mworld\033[0m ansi colors\t\033[32mwrap\033[0m\t?",
        "this is \033[7mreversed\033[7m!",
        "\033[43;30mYellow\033[0m \033[45;37mMagenta"
    ])


def test_ansi_colors_scroll():
    reference = 'ansi_colors_scroll.png'
    device = dummy()
    term = terminal(device, font=test_font, word_wrap=True, animate=False)

    assert_text(device, term, reference, [
        "hello \033[31mworld\033[0m ansi colors\t\033[32mwrap\033[0m\t?",
        "this is \033[7mreversed\033[7m!",
        "\033[43;30mYellow\033[0m \033[44;37mBlue abcdefg hijklmn",
        "\033[41;30mRed\033[0m \033[42;37mGreen"
    ])


def test_accented_charset():
    reference = 'accented_charset.png'
    unicode_font = get_reference_font('DejaVuSans.ttf')
    device = dummy()
    term = terminal(device, font=unicode_font, word_wrap=False, animate=False,
        color="blue", bgcolor="white")

    assert_text(device, term, reference, [
        u"\033[31mFußgängerunterführungen\033[0m Текст на русском"
    ])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2024 Richard Hull and contributors
# See LICENSE.rst for details.

from helpers import test_font


def primitives(device, draw):
    padding = 2
    shape_width = 20
    top = padding
    bottom = device.height - padding - 1
    draw.rectangle(device.bounding_box, outline="white", fill="black")
    x = padding
    draw.ellipse((x, top, x + shape_width, bottom), outline="red", fill="black")
    x += shape_width + padding
    draw.rectangle((x, top, x + shape_width, bottom), outline="blue", fill="black")
    x += shape_width + padding
    draw.polygon([(x, bottom), (x + shape_width / 2, top), (x + shape_width, bottom)], outline="green", fill="black")
    x += shape_width + padding
    draw.line((x, bottom, x + shape_width, top), fill="yellow")
    draw.line((x, top, x + shape_width, bottom), fill="yellow")
    x += shape_width + padding
    draw.text((x, top), 'Hello', font=test_font, fill="cyan")
    draw.text((x, top + 20), 'World!', font=test_font, fill="purple")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.

import pytest

from PIL import Image
from luma.core.device import dummy
from luma.core.virtual import sevensegment

from helpers import get_reference_image, assert_identical_image

_DIGITS = {
    ' ': 0x00,
    '-': 0x01,
    '_': 0x08,
    '\'': 0x02,
    '0': 0x7e,
    '1': 0x30,
    '2': 0x6d,
    '3': 0x79,
    '4': 0x33,
    '5': 0x5b,
    '6': 0x5f,
    '7': 0x70,
    '8': 0x7f,
    '9': 0x7b,
    'a': 0x7d,
    'b': 0x1f,
    'c': 0x0d,
    'd': 0x3d,
    'e': 0x6f,
    'f': 0x47,
    'g': 0x7b,
    'h': 0x17,
    'i': 0x10,
    'j': 0x18,
    # 'k': cant represent
    'l': 0x06,
    # 'm': cant represent
    'n': 0x15,
    'o': 0x1d,
    'p': 0x67,
    'q': 0x73,
    'r': 0x05,
    's': 0x5b,
    't': 0x0f,
    'u': 0x1c,
    'v': 0x1c,
    # 'w': cant represent
    # 'x': cant represent
    'y': 0x3b,
    'z': 0x6d,
    'A': 0x77,
    'B': 0x7f,
    'C': 0x4e,
    'D': 0x7e,
    'E': 0x4f,
    'F': 0x47,
    'G': 0x5e,
    'H': 0x37,
    'I': 0x30,
    'J': 0x38,
    # 'K': cant represent
    'L': 0x0e,
    # 'M': cant represent
    'N': 0x76,
    'O': 0x7e,
    'P': 0x67,
    'Q': 0x73,
    'R': 0x46,
    'S': 0x5b,
    'T': 0x0f,
    'U': 0x3e,
    'V': 0x3e,
    # 'W': cant represent
    # 'X': cant represent
    'Y': 0x3b,
    'Z': 0x6d,
    ',': 0x80,
    '.': 0x80
}


def dot_muncher(text, notfound="_"):
    undefined = _DIGITS[notfound]
    iterator = iter(text)
    try:
        last = _DIGITS.get(next(iterator), undefined)
    except StopIteration:
        # Catch exception from next() in py3.7 and return normally
        return

    try:
        while True:
            curr = _DIGITS.get(next(iterator), undefined)

            if curr == 0x80:
                yield curr | last
            elif last != 0x80:
                yield last

            last = curr
    except StopIteration:
        if last != 0x80:
            yield last


def test_init():
    device = dummy(width=24, height=8, mode="1")
    sevensegment(device, segment_mapper=dot_muncher)
    assert device.image == Image.new("1", (24, 8))


def test_overflow():
    device = dummy(width=24, height=8, mode="1")
    seg = sevensegment(device, segment_mapper=dot_muncher)
    with pytest.raises(OverflowError) as ex:
        seg.text = "This is too big to fit in 3x8 seven-segment displays"
    assert str(ex.value) == "Device's capabilities insufficient for value 'This is too big to fit in 3x8 seven-segment displays'"


def test_setter_getter():
    img_path = get_reference_image('golden_ratio.png')

    with open(img_path, 'rb') as img:
        reference = Image.open(img)
        device = dummy(width=24, height=8)
        seg = sevensegment(device, segment_mapper=dot_muncher)
        seg.text = "1.61803398875"
        assert str(seg.text) == "1.61803398875"

        assert_identical_image(reference, device.image, img_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Richard Hull and contributors
# See LICENSE.rst for details.


from unittest.mock import patch

from luma.core.device import dummy


def test_persist():
    dev = dummy()
    assert dev.persist is False
    with patch.object(dev, 'hide') as mock:
        dev.cleanup()
    mock.assert_called_once_with()
    dev = dummy()
    dev.persist = True
    with patch.object(dev, 'hide') as mock:
        dev.cleanup()
    mock.assert_not_called()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.

from luma.core.device import dummy
from luma.core.legacy import text, textsize, show_message
from luma.core.legacy.font import proportional, CP437_FONT, LCD_FONT

from unittest.mock import Mock, call


def test_textsize():
    """
    The bounding box of the text, as drawn in the specified font, is correctly
    calculated.
    """
    assert textsize("Hello world") == (88, 8)
    assert textsize("Hello world", font=proportional(CP437_FONT)) == (71, 8)


def test_text_space():
    """
    Draw a space character.
    """
    draw = Mock(unsafe=True)
    text(draw, (2, 2), " ", fill="white")
    draw.point.assert_not_called()


def test_text_char():
    """
    Draw a text character.
    """
    draw = Mock(unsafe=True)
    text(draw, (2, 2), "L", font=LCD_FONT, fill="white")
    draw.point.assert_has_calls([
        call((2, 2), fill='white'),
        call((2, 3), fill='white'),
        call((2, 4), fill='white'),
        call((2, 5), fill='white'),
        call((2, 6), fill='white'),
        call((2, 7), fill='white'),
        call((2, 8), fill='white'),
        call((3, 8), fill='white'),
        call((4, 8), fill='white'),
        call((5, 8), fill='white'),
        call((6, 8), fill='white')
    ])


def test_show_message():
    """
    Scroll a message right-to-left across the devices display.
    """
    device = dummy()
    show_message(device, 'text', scroll_delay=0.0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:module:`luma.core.bitmap_font` module.
"""

import os
from pathlib import Path

import cbor2
import pytest
from PIL import Image, ImageDraw
from luma.core import bitmap_font
from luma.core.bitmap_font import embedded_fonts

from helpers import get_reference_pillow_font, get_reference_file


FONTDATA = {
    'metrics': [
        # A00 ENGLISH_JAPANESE 5x8 METRICS
        {
            'name': 'A00',
            'index': range(16, 256),
            'xwidth': 5,
            'cell_size': (5, 10),
            'glyph_size': (5, 8),
            'table_size': (800, 20)
        },
        # A02 ENGLISH_EUROPEAN 5x8 METRICS
        {
            'name': 'A02',
            'index': range(16, 256),
            'xwidth': 5,
            'cell_size': (5, 10),
            'glyph_size': (5, 8),
            'table_size': (800, 20)
        }
    ],
    'mappings': [
        {
            0x00a5:   0x5c,  # ¥ YEN SIGN
            0x2192:   0x7e,  # → RIGHTWARDS ARROW
            0x2190:   0x7f,  # ← LEFTWARDS ARROW
            0x300c:   0xa2,  # 「 LEFT CORNER BRACKET
            0x300d:   0xa3,  # 」 RIGHT CORNER BRACKET
            0x30a1:   0xa7,  # ァ KATAKANA LETTER SMALL A
            0x30a3:   0xa8,  # ィ KATAKANA LETTER SMALL I
            0x30a5:   0xa9,  # ゥ KATAKANA LETTER SMALL U
            0x30a7:   0xaa,  # ェ KATAKANA LETTER SMALL E
            0x30a9:   0xab,  # ォ KATAKANA LETTER SMALL O
            0x30e3:   0xac,  # ャ KATAKANA LETTER SMALL YA
            0x30e5:   0xad,  # ュ KATAKANA LETTER SMALL YU
            0x30e7:   0xae,  # ョ KATAKANA LETTER SMALL YO
            0x30c3:   0xaf,  # ッ KATAKANA LETTER SMALL TU
            0x00f7:   0xfd,  # ÷ DIVISION SIGN
            0x25ae:   0xff,  # ▮ BLACK VERTICAL RECTANGLE
        },
        {   # A02 ENGLISH_EUROPEAN CHARACTER FONT
            # Contains no 5x10 fonts
            0x25b6:   0x10,  # ▶ BLACK RIGHT-POINTING TRIANGLE
            0x25c0:   0x11,  # ◀ BLACK LEFT-POINTING TRIANGLE
            0x201c:   0x12,  # “ LEFT DOUBLE QUOTATION MARK
            0x201d:   0x13,  # ” RIGHT DOUBLE QUOTATION MARK
            0x23eb:   0x14,  # ⏫ BLACK UP-POINTING DOUBLE TRIANGLE
            0x23ec:   0x15,  # ⏬ BLACK DOWN-POINTING DOUBLE TRIANGLE
            0x25cf:   0x16,  # ● BLACK CIRCLE
            0x00bc:   0xbc,  # ¼ VULGAR FRACTION ONE QUARTER
            0x00bd:   0xbd,  # ½ VULGAR FRACTION ONE HALF
            0x00be:   0xbe,  # ¾ VULGAR FRACTION THREE QUARTERS
            0x00bf:   0xbf,  # ¿ INVERTED QUESTION MARK
            0x00c0:   0xc0,  # À LATIN CAPITAL LETTER A WITH GRAVE
            0x00c1:   0xc1,  # Á LATIN CAPITAL LETTER A WITH ACUTE
            0x00c2:   0xc2,  # Â LATIN CAPITAL LETTER A WITH CIRCUMFLEX
            0x00c3:   0xc3,  # Ã LATIN CAPITAL LETTER A WITH TILDE
            0x00c4:   0xc4,  # Ä LATIN CAPITAL LETTER A WITH DIAERESIS
            0x00c5:   0xc5,  # Å LATIN CAPITAL LETTER A WITH RING ABOVE
            0x00c6:   0xc6,  # Æ LATIN CAPITAL LETTER AE
            0x0393:   0x92,  # Γ GREEK CAPITAL LETTER GAMMA
            0x0413:   0x92,  # Г CYRILLIC CAPITAL LETTER GHE
        }
    ],
    'fonts': [
        # A00 ENGLISH_JAPANESE CHARACTER FONT
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x14\xa2a\x8c\x12\x00\x00\x00\x00q\x1d\xf1|\xdfs\x80\x01\x01\x0es\xbc\xee\x7f\xee\x8b\x8f\x18F.\xf3\xbc\xff\xc61\x8c\x7f\xc8\xb8\x80@ \x00\x80\xc0\x81\x05\x06\x00\x00\x00\x00\x04\x00\x00\x00\x00" \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x14\xa7\xe6D!\x08@\x00\x01\x8b"#A\x11\x8cX\xc2\x00\x91\x8cc\x19B\x11\x89\x05(n1\x8cc\x02F1\x8cC\x05\t@  \x00\x81/\x80\x01\x02\x00\x00\x00\x00\x04\x00\x00\x00\x00B\x10\x84\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00\x03\xe0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x15\xfa\n\x88@\xaa@\x00\x02\x99\x02Ez\x01\x8cX\xc4|A\x0cc\x08\xc2\x10\x89\x05HW1\x8cc\x02F1TE\x0f\x8a \x13\xac\xe6\xb9\x11\xb3\r"j\xce\xf3l\xeeF1\x8c~B\x10H\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00\x00?\x11>$\x03\xc0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\xa7\x11\x00@\x9d\xf0|\x04\xa9\x04)\x07\xc2s\xc0\x08\x00"l}\x08\xfb\xd7\xf9\x05\x88V\xb1\xf4|\xe2F5"\x89\x02\x08\x00\x00s\t\xc7\x91\xc9\x05BW1\x8c\xf3\x04F5TD\x82\x0b\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08@3\xe1\'\xc9\xff\xb8U\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xf2\xa2\xa0@\xaaF\x00\x08\xc9\x08\x1f\x86$\x88X\xc4|D\xaf\xe3\x08\xc2\x11\x89\x05HFq\x85h\x12F5Q\x11\x0f\x88\x00\x03\xe3\x08\xfd\x0f\x89\x05\x82V1\xf3\xe0\xe4F5#\xc8B\x10H\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\x00H0&dHd\x8b\xd5\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xafN@!\x08B\x01\x90\x89\x11\x11F$\x88\x98B\x00\x80\xacc\x19B\x11\x89%(F1\x84\xa4\x12EU\x89!\x02\x08\x00\x04c\x18\xc1\x01\x89%BF1\x80`\x14\xcdUPPB\x10\x84\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x05\x00D\x00D\xa0H\xa5\x08A\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\xa2\r\xa0\x12\x00\x04\x01\x80s\xbe\xe19\xc4s\x00\x81\x01\x04t|\xee~\x0f\x8b\x99\x1f\xc6.\x83c\xe28\x8a\x89?\xc28\x1f\x03\xfc\xe7\xb9\x0e\x8b\x99\'F.\x80a\xe34\x8a\x8b\xbe" \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\x01\xc2\x00\x88!\xbf$\x7f\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\xc2@\t\x04\x02\x00\xa0\x01\x00\x00\x80\xe4\x10\x00 \x81\x00\x00\x80\x00\x00\x04\x00\x0e\x90\x00\x00\x00\x9c\x02\x80\x00\x00\x00\x00\x04\x00!\xca\x00\x00\x00+\xe0\xf8\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00E\xff\xff\xff{\xff\xfc\x7f\xf1\x7f*\x04\x11\xdf\xf8\x89\x0f\xa3\xffq\x03\xf4;\xe0\x91!\xff\xe2T\x00\x00\x00\x00\x00\x00\x81B \x00\x00\x1c\x00\x02\x00\x00\x02\x00\x10\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01I\x12\t$L\x82\xa0\x851I+\xf4|\x01\x10\x85\xf0\xd0\x81\x02\x02\x8f\x88?\x95!\x18\x81\x1cK\x9c\xe8\xbc\xcf>\x8c\x87r\xce\xb3b\x07E\x1f\x8c}\xff\x80\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf9\x99\x12\x19?\x88\x82\xac\x89I\xaf\xeaF\x10\n \xa3\x00\x88\x81t\x15\xf4\x8b\xe1\x95#\x18\x84\x00\xa8c\x08\xd11 \x85J#1\xcc\xfe\xb8\xc4\x8aTH\x8a\xfc\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01(\x12)$\x08\x82 \x91\x01\x19\x02E\x10\x04p\xa3\x00\x86\xaa\x04H\x85\x08!\x15e\x10\x84\x00\x93\xfc\xc8\xca1 \x04\n\xf21\x8ccX\xc5\n$~\xff\x80\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x08"I$\x10\x82A)\x02\x11\x04D \n\xa9#\x01\x06\xa4w\xd4\x84\x08"%i\x11\x08\x00\x94c\x19\xc61\xa0\x04\x07"1\x8cc\xa5F\nTH\x98\x90\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x08O\x8aDa>\x8eD\xecb\x08\x84C\xf0""\xf6\x00\x82\x08`t\x7f\xe4E\xb1\xf2p\x00k\xfc\xee\xbb\xcf@\x04\x02>.\xf3\xdc\r\xcf\xf3\x8b\xc9\x18\x80\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 \x08\x02\x01\x00\x04\x00\x00\x00\x80@\x004\x00\x00@\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 \x08\x02\x01\x00$\x00\x00\x00\x80@\x00\x00\x00\x00@\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 \x08\x02\x0e\x00\x18\x00\x00\x00\x80@\x00\x00\x00\x03\x80\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',

        # A02 WESTERN EUROPEAN CHARACTER FONT
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\xc0\x00(\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00@\x93\xb2|\x01!\x00\x01 \x00\x01\x14\xa2a\x8c\x12\x00\x00\x00\x00q\x1d\xf1|\xdfs\x80\x01\x01\x0eq<\xee\x7f\xee\x8b\x8f\x18F.\xf3\xbc\xef\xc61\x8c~\xe08\x80@ \x00\x80\xc0\x81\x05\x06\x00\x00\x00\x00\x04\x00\x00\x00\x00" \x04\xf9k\xe8\x91\xff\x8cb\n\xe2.\x01>\x0f\x80\xa0#\x80`\x00\x0e\xd9\x08`D\x86\x17\xdc\t?\xe4a\xa4\x979\xc1q\x08B\x10\x9f\x01\x14\xa7\xe6D!\x08@\x00\x01\x8b"#A\x11\x8cX\xc2\x00\x91\x8a\xa3\x19B\x11\x89\x05(n1\x8cc\x12F1\x8cB\x88\t@  \x00\x81 \x80\x01\x02\x00\x00\x00\x00\x04\x00\x00\x00\x00B\x10\n\x89j\x18\xc4\xb1\x8ccZ\xa21\x01\xa2\x08\x00\xe1t\\\x90(\x11\xd8\x1c\x88\xa8\x89,BZ\xc6(s\xb7/\x93\xe5\xa9\x04\x84\x08\x8e\x01\x15\xfa\n\x88@\xaa@\x00\x02\x99\x02Ez\x01\x8cX\xc4|A\x0cc\x08\xc2\x10\x89\x05HW1\x8cc\x02F1TD\x84\n \x13\xac\xe6\xb9\x0f\xb1\r"j\xce\xf3l\xeeF1\x8c~B\x10\x11\x82j\x19\xc4\xb1\x8ccZ\xa2%Ia\xf4<\xaetbE\xfd\xd1\xd8(\x87|\x84%^\xaa\xc6\xac\x7f\x80\x00\x03\xe9!?\xf2\x11\xce\x01\x00\xa7\x11\x00@\x9d\xf0|\x04\xa9\x04)\x07\xc2s\xc0\x08\x00"l}\x08\xfb\xd7\xf9\x05\x88V\xb1\xf4|\xe2F5"\x88\x82\x08\x00\x00s\t\xc7\x91\xcb\x05BW1\x8c\xf3\x04F1TD\x82\t\xb1\xf4\\j\xcc\xb1T_Z\xbb+\xa9`\xa2H\xb4w\xe2\xaa\xfe\x11\xd9)\xc5\x10\n\xfd\xe3N\xbe s\x80\x02\x7f\xff%D\x81!\xc4\x00\x01\xf2\xa2\xa0@\xaaF\x00\x08\xc9\x08\x1f\x86$\x88X\xc4|D\xaf\xe3\x08\xc2\x11\x89\x05HFq\x85h\x12F5Q\x10\x81\x08\x00\x03\xe3\x08\xfd\x0f\x89\x05\x82V1\xf3\xe0\xe4F5#\xc8B\x12Q\x8f\xea\x1c\xd4\xb1$CZ\xa6\xa1\x91 \xa4K\xa4\xfcc\x1d}\x91\xd9*\x87|\x84%^\xaa\x96`a\x80\x079\xc8#\x88@\x03\xe4\x00\x00\xafN@!\x08B\x01\x90\x89\x11\x11F$\x88\x98B\x00\x80\xacc\x19B\x11\x89%(F1\x84\xa5\x12EU\x89 \x80\x88\x00\x04c\x18\xc1\x01\x89%BV1\x80`\x14\xcdUPPB\x10\x1f\x8cj\x18\xe6\xb1G\xc3_\xa6\xb1\x97 \xa8Kd$U\x10:1\xd9\x1c\x98\x90\x92\xa4@Z\xa6\xa0@\x80\x0f\x90\x04!\x00\x0f\xfc\x00\x01\x00\xa2\r\xa0\x12\x00\x04\x01\x80s\xbe\xe19\xc4s\x00\x81\x01\x04t|\xee~\x0f\x8b\x99\x1f\xc6.\x83b\xe28\x8a\x89>\xe08\x1f\x03\xfc\xe7\xb9\x0e\x8b\x99\'V.\x80a\xe34\x8a\x8b\xbe" \x00\xf4k\xe8\xc51\x80C\xf0\xbb.o!?\xb0b\x03\xb6\xe0\x11\xd1\xd9\t`\x10\x8cG\xfe\tG\xe0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00`\x19\xce\x00\x00\x02\x00\x08\xc7\x00@\x88\xd5\x10\x0e@\x88\x04\x08\x80\x03P"5@\x02\x04E\x0b\x00@\x88\xd0\x10\x00@\x88\x04\x08\x80\x03P \x00\x00\x02\x04@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x91$)E\xe0\x06\x1c\tI\x04!\x15 (\xf1!\x14\xa2\x11Jt\x88EH\x00q\x08\xa0\x11\x06!\x15%(\x00!\x14\xa2\x11J\xa4\x88B5@\x11\x08\xa5\t\x8a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x91\x08\xceF`\x02#JS\x00!\x00\x02\x11\x90\x00\x00\x00\x00\x00H\x1c\xe0\x01\xd1$b\x08\xc5\xc9\x00\x00\x00\x13N\x00\x00\x00\x00\x00@\x00\x05H\x04 \x00\x00\x10\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x97\xd0(F`R"\xa5)$R\x9c\xe5:\x90\xff\xff\xf79\xceLc\x17:*tc\x18\xa9)s\x9c\xe78\xb0s\x9c\xe2\x10\x84\xa5\x9c\xe0\x01\xc0tc\x18\xc4\xd1\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00a=\xc9M\xec\x8f"[Wh\x8cc\x18\xc6\xf1\x84!\x02\x10\x84\xeec\x18\xc6$\xacc\x18\x91.\x08B\x10\x85\xf1\x8cc\x161\x8c\x16c\x17:?\xacc\x18\xc4\xb1\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x0b\xf4l\xa8\x1c\xa5\x04\xb0\xff\xff\xff\xff\x8e\xf7\xbd\xe2\x10\x84Mc\x18\xc6*tc\x18\x91\xc9{\xde\xf7\xbe\x8e\xff\xff\xf2\x10\x84|c\x18\xc6 tc\x18\xbc\xcf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\t@`\xa8\x01G\x88\xf1\x8cc\x18\xc6\x82\x84!\x02\x10\x84L\xe3\x18\xc61$c\x18\x91\t\x8cc\x18\xc6\xa4\x84!\x02\x10\x84\x8cc\x18\xc6$$\xe79\x84\x81\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\xc0\x01\xc0`P>\x01\x1c.\x8cc\x18\xc6\xe6\xff\xff\xf79\xcet\\\xe79\xc0s\x9c\xe7\x13\x96{\xde\xf7\xbdLs\x9c\xe79\xcet\\\xe79\xc0CZ\xd6\xb9\xce\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    ]
}


def make_sprite_table(fnt):
    """
    Generate a page filled with glyphs from character value 16-255 using the
    provided font

    :param fnt: The font to use to draw the page
    :type fnt: `PIL.ImageFont`
    :return: Image containing the drawn glyphs
    :rtype: `PIL.Image`
    """
    img = Image.new('1', (80, 120), 0)
    drw = ImageDraw.Draw(img)
    for i in range(240):
        drw.text(((i % 16) * 5, (i // 16) * 8), chr(i + 16), font=fnt, fill='white', spacing=0)
    return img


@pytest.fixture()
def bm_font(request):
    """
    Fixture which loads a ``bitmap_font`` persists it to disk
    The fixture removes the file when it is finished.
    """
    filename = get_reference_file(Path('font').joinpath('hd44780a02.pbm'))
    bmf = bitmap_font.load_sprite_table(filename, range(16, 256), 5, (5, 8),
        (5, 8), FONTDATA['mappings'][1])
    bmf.save('test.bmf')

    def tear_down():
        os.remove('test.bmf')
    request.addfinalizer(tear_down)

    return bmf


@pytest.fixture()
def load_all_embedded(request):
    """
    Fixture which loads the two fonts contained within the test FONTDATA information
    """
    fnt_cnt = len(FONTDATA['metrics'])
    bmfs = [None] * fnt_cnt

    for i in range(fnt_cnt):
        m = FONTDATA['metrics'][i]
        sp_data = FONTDATA['fonts'][i]
        mappings = FONTDATA['mappings'][i]
        sprite_table = Image.frombytes('1', m['table_size'], sp_data)
        bmfs[i] = bitmap_font.load_sprite_table(sprite_table, m['index'], m['xwidth'],
            m['glyph_size'], m['cell_size'], mappings)
        sprite_table.close()
    return bmfs


def test_load_from_pillow_font():
    """
    Test the loading of a pillow font from disk by loading the font from bitmap_font
    and PIL.ImageFont, rendering a page of glyphs which each and testing to make
    sure the results are the same.
    """
    fnt = get_reference_pillow_font('hd44780a02.pil')
    filename = get_reference_file(Path('font').joinpath('hd44780a02.pil'))
    bmf = bitmap_font.load_pillow_font(filename)

    img1 = make_sprite_table(fnt)
    img2 = make_sprite_table(bmf)

    assert img1 == img2


def test_load_from_pillow_exceptions():
    """
    Test that exceptions are thrown as appropriate if bitmap_font is asked to
    load a pillow font that is not a PIL.ImageFont file, is damaged or does not
    include the required font metrics data.
    """

    with pytest.raises(SyntaxError) as ex:
        filename = get_reference_file(Path('font').joinpath('hd44780a02.pbm'))
        bitmap_font.load_pillow_font(filename, FONTDATA['mappings'][1])
    assert str(ex.value) == "Not a PIL.ImageFont file"

    with pytest.raises(SyntaxError) as ex:
        filename = get_reference_file(Path('font').joinpath('hd44780a02_nodata.pil'))
        bitmap_font.load_pillow_font(filename, FONTDATA['mappings'][1])
    assert str(ex.value) == "PIL.ImageFont file missing metric data"

    with pytest.raises(SyntaxError) as ex:
        filename = get_reference_file(Path('font').joinpath('hd44780a02_incomplete.pil'))
        bitmap_font.load_pillow_font(filename, FONTDATA['mappings'][1])
    assert str(ex.value) == "PIL.ImageFont file metric data incomplete"

    filename = get_reference_file(Path('font').joinpath('wrong_mode.pil'))
    with pytest.raises(OSError) as ex:
        bitmap_font.load_pillow_font(filename)
    assert str(ex.value) == 'cannot find glyph data file'


def test_mapping():
    """
    Test to make sure that values that have unicode mappings work correctly
    """
    fnt = get_reference_pillow_font('hd44780a02.pil')
    filename = get_reference_file(Path('font').joinpath('hd44780a02.pil'))
    bmf = bitmap_font.load_pillow_font(filename, FONTDATA['mappings'][1])

    size = bmf.getsize('\u0010')
    img1 = Image.new('1', size, 0)
    img2 = Image.new('1', (5, 8), 0)

    drw = ImageDraw.Draw(img1)
    drw.text((0, 0), '\u0010', font=fnt, fill='white')
    drw = ImageDraw.Draw(img2)
    drw.text((0, 0), '\u25b6', font=bmf, fill='white')

    assert img1 == img2


def test_load_sprite_table():
    """
    Test loading a font from a sprite_table
    """
    fnt = get_reference_pillow_font('hd44780a02.pil')
    filename = get_reference_file(Path('font').joinpath('hd44780a02.pbm'))
    bmf = bitmap_font.load_sprite_table(filename, range(16, 256), 5, (5, 8), (5, 8), FONTDATA['mappings'][1])

    img1 = make_sprite_table(fnt)
    img2 = make_sprite_table(bmf)

    assert img1 == img2


def test_load_sprite_table_exceptions_1():
    """
    Test that exceptions are thrown as appropriate if bitmap_font is asked to
    load from a sprite table from a filename that does not exist, is not a
    PIL.Image file, or is damaged.
    """
    with pytest.raises(FileNotFoundError) as ex:
        filename = 'badfile'
        bitmap_font.load_sprite_table(filename, range(16, 256), 5, (5, 8), (5, 8), FONTDATA['mappings'][1])
    assert ex.value.filename == filename


def test_load_sprite_table_exceptions_2():
    """
    Test that exceptions are thrown as appropriate if bitmap_font is asked to
    load from a sprite table from a filename that does not exist, is not a
    PIL.Image file, or is damaged.
    """
    with pytest.raises(ValueError) as ex:
        filename = get_reference_file(Path('font').joinpath('hd44780a02.pil'))
        bitmap_font.load_sprite_table(filename, range(16, 256), 5, (5, 8), (5, 8), FONTDATA['mappings'][1])
    assert str(ex.value) == f'File {filename} not a valid sprite table'

    with pytest.raises(ValueError) as ex:
        bitmap_font.load_sprite_table(1, range(16, 256), 5, (5, 8), (5, 8), FONTDATA['mappings'][1])
    assert str(ex.value) == 'Provided image is not an instance of PIL.Image'


def test_dumps_loads_saves_load(bm_font):
    """
    Test which verifies the loading and restoring of bitmap_fonts
    """
    bmf = bm_font

    data = bmf.dumps()
    bmf2 = bitmap_font.loads(data)
    bmf3 = bitmap_font.load('test.bmf')

    img1 = make_sprite_table(bmf)
    img2 = make_sprite_table(bmf2)
    img3 = make_sprite_table(bmf3)

    assert img1 == img2
    assert img1 == img3

    with pytest.raises(SyntaxError) as ex:
        filename = get_reference_file(Path('font').joinpath('hd44780a02.pil'))
        bitmap_font.load(filename)
    assert str(ex.value) == 'Not a luma.core.bitmap_font file'


def test_loads_exception():
    """
    Exception is thrown when attempting to load font data that is incomplete.
    """
    data = cbor2.dumps({'xwidth': 5, 'glyph_size': (5, 8)})
    with pytest.raises(Exception) as ex:
        bitmap_font.loads(data)
    assert str(ex.value) == 'Cannot parse fontdata. It is invalid.'


def test_combine(load_all_embedded):
    """
    Fonts can be combined successfully.
    """
    bmfs = load_all_embedded

    # Demonstrate that \u25b6 and \u25c0 are not in bmfs[0]
    img = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img)
    drw.text(((0, 0)), '\u00a5\u25b6', font=bmfs[0], fill='white', spacing=0)
    assert sum(img.crop((5, 0, 9, 7)).tobytes()) == 0
    drw.text(((0, 0)), '\u00a5\u25c0', font=bmfs[0], fill='white', spacing=0)
    assert sum(img.crop((5, 0, 9, 7)).tobytes()) == 0

    # Combine a couple of characters (one new, one conflicting)
    bmfs[0].combine(bmfs[1], '\u25b6\U000f8041')
    img1 = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img1)
    drw.text(((0, 0)), '\u00a5\u25b6', font=bmfs[0], fill='white', spacing=0)

    img2 = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img2)
    drw.text(((0, 0)), '\u00a5', font=bmfs[0], fill='white', spacing=0)
    drw.text(((5, 0)), '\u25b6', font=bmfs[1], fill='white', spacing=0)

    assert img1 == img2

    img2 = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img2)
    drw.text(((0, 0)), '\u00a5', font=bmfs[1], fill='white', spacing=0)
    drw.text(((5, 0)), '\u25b6', font=bmfs[0], fill='white', spacing=0)

    assert img1 != img2

    # Combine entire font
    bmfs[0].combine(bmfs[1])
    img1 = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img1)
    drw.text(((0, 0)), '\u00a5\u25c0', font=bmfs[0], fill='white', spacing=0)

    img2 = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img2)
    drw.text(((0, 0)), '\u00a5', font=bmfs[0], fill='white', spacing=0)
    drw.text(((5, 0)), '\u25c0', font=bmfs[1], fill='white', spacing=0)

    assert img1 == img2

    with pytest.raises(ValueError) as ex:
        c = '\u0000'
        bmfs[0].combine(bmfs[1], c)
    assert str(ex.value) == f'{c} is not a valid character within the source font'


def test_embedded_font(load_all_embedded):
    """
    Tests the embedded_fonts class.  Loads FONTDATA.  Changes current font
    by name and number.  Combines two fonts and verifies that combined
    characters get rendered correctly.
    """
    bmfs = load_all_embedded
    ebf = embedded_fonts(FONTDATA)

    assert ebf.current == ebf.font_by_number[0]
    ebf.current = 'A02'
    assert ebf.current == ebf.font_by_number[1]
    a02fnt = ebf.current
    ebf.current = 0
    assert ebf.current == ebf.font_by_number[0]

    ebf.combine(a02fnt)

    img1 = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img1)
    drw.text(((0, 0)), '\u00a5\u25c0', font=ebf, fill='white', spacing=0)

    img2 = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img2)
    drw.text(((0, 0)), '\u00a5', font=bmfs[0], fill='white', spacing=0)
    drw.text(((5, 0)), '\u25c0', font=bmfs[1], fill='white', spacing=0)

    assert img1 == img2


def test_embedded_font_exceptions():
    """
    Tests exceptions when attempting to select a font that does not exist
    within the ``embedded_fonts`` object
    """
    ebf = embedded_fonts(FONTDATA)

    with pytest.raises(ValueError) as ex:
        ebf.current = 2
    assert str(ex.value) == 'No font with index 2'

    with pytest.raises(ValueError) as ex:
        ebf.current = 'BAD'
    assert str(ex.value) == 'No font with name BAD'

    with pytest.raises(TypeError) as ex:
        ebf.current = []
    assert str(ex.value) == f'Expected int or str.  Received {type([])}'


def test_metrics(load_all_embedded):
    """
    ``bitmap_font`` correctly handles fonts that have characters within
    the font that have offsets.
    """
    bmfs = load_all_embedded

    img1 = Image.new('1', (10, 8), 0)
    drw = ImageDraw.Draw(img1)
    drw.text(((0, 0)), 'ij', font=bmfs[0], fill='white', spacing=0)
    size = bmfs[0].getsize('ij')
    assert size == (10, 8)

    img1d = Image.new('1', (10, 10), 0)
    drw = ImageDraw.Draw(img1d)
    drw.text(((0, 0)), 'i', font=bmfs[0], fill='white', spacing=0)
    drw.text(((5, 2)), 'j', font=bmfs[0], fill='white', spacing=0)

    # Make j descend 2 pixels below the baseline
    j = bmfs[0]._lookup(ord('j'))
    dst = bmfs[0].metrics[j]['dst']
    dst[1] = dst[1] + 2
    dst[3] = dst[3] + 2
    bmfs[0]._calculate_font_size()

    size = bmfs[0].getsize('ij')
    assert size == (10, 10)

    img2 = Image.new('1', size, 0)
    drw = ImageDraw.Draw(img2)
    drw.text(((0, 0)), 'ij', font=bmfs[0], fill='white', spacing=0)

    assert img1 != img2
    assert img1d == img2

    data = bmfs[0].dumps()
    restore = bitmap_font.loads(data)

    img2 = Image.new('1', size, 0)
    drw = ImageDraw.Draw(img2)
    drw.text(((0, 0)), 'ij', font=restore, fill='white', spacing=0)
    assert img1d == img2


def test_glyph_index(load_all_embedded):
    """
    ``glyph_table`` contains expected values
    """
    bmfs = load_all_embedded

    # Test space 'A' and 'a' characters
    for i in [0x20, 0x41, 0x61]:
        img = Image.new('1', (5, 8), 0)
        drw = ImageDraw.Draw(img)
        drw.text((0, 0), chr(i), font=bmfs[0], fill='white')
        assert bmfs[0].glyph_index[img.tobytes()] == i, f'{i} does not match its glyph_index value'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:mod:`luma.core.ansi_color` module.
"""

import pytest

from luma.core import ansi_color


def test_parse_str_no_escape():
    gen = ansi_color.parse_str("hello world")
    assert next(gen) == ["putch", "h"]
    assert next(gen) == ["putch", "e"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "o"]
    assert next(gen) == ["putch", " "]
    assert next(gen) == ["putch", "w"]
    assert next(gen) == ["putch", "o"]
    assert next(gen) == ["putch", "r"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "d"]

    with pytest.raises(StopIteration):
        next(gen)


def test_parse_str_valid_ansi_colors():
    gen = ansi_color.parse_str("hello \033[31mworld\33[0m")
    assert next(gen) == ["putch", "h"]
    assert next(gen) == ["putch", "e"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "o"]
    assert next(gen) == ["putch", " "]
    assert next(gen) == ["foreground_color", "red"]
    assert next(gen) == ["putch", "w"]
    assert next(gen) == ["putch", "o"]
    assert next(gen) == ["putch", "r"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "d"]
    assert next(gen) == ["reset"]

    with pytest.raises(StopIteration):
        next(gen)


def test_parse_str_valid_ansi_colors_extended_codeset():
    gen = ansi_color.parse_str(u"á \033[31mFußgänger Текст на\033[0m")
    assert next(gen) == ["putch", u"á"]
    assert next(gen) == ["putch", " "]
    assert next(gen) == ["foreground_color", "red"]
    assert next(gen) == ["putch", "F"]
    assert next(gen) == ["putch", "u"]
    assert next(gen) == ["putch", u"ß"]
    assert next(gen) == ["putch", "g"]
    assert next(gen) == ["putch", u"ä"]
    assert next(gen) == ["putch", "n"]
    assert next(gen) == ["putch", "g"]
    assert next(gen) == ["putch", "e"]
    assert next(gen) == ["putch", "r"]
    assert next(gen) == ["putch", " "]
    assert next(gen) == ["putch", u"Т"]
    assert next(gen) == ["putch", u"е"]
    assert next(gen) == ["putch", u"к"]
    assert next(gen) == ["putch", u"с"]
    assert next(gen) == ["putch", u"т"]
    assert next(gen) == ["putch", " "]
    assert next(gen) == ["putch", u"н"]
    assert next(gen) == ["putch", u"а"]
    assert next(gen) == ["reset"]

    with pytest.raises(StopIteration):
        next(gen)


def test_parse_str_multiple_ansi_colors():
    gen = ansi_color.parse_str("hello \033[32;46mworld\33[7;0m")
    assert next(gen) == ["putch", "h"]
    assert next(gen) == ["putch", "e"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "o"]
    assert next(gen) == ["putch", " "]
    assert next(gen) == ["foreground_color", "green"]
    assert next(gen) == ["background_color", "cyan"]
    assert next(gen) == ["putch", "w"]
    assert next(gen) == ["putch", "o"]
    assert next(gen) == ["putch", "r"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "d"]
    assert next(gen) == ["reverse_colors"]
    assert next(gen) == ["reset"]

    with pytest.raises(StopIteration):
        next(gen)


def test_parse_str_unknown_ansi_colors_ignored():
    gen = ansi_color.parse_str("hello \033[27mworld")
    assert next(gen) == ["putch", "h"]
    assert next(gen) == ["putch", "e"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "o"]
    assert next(gen) == ["putch", " "]
    assert next(gen) == ["putch", "w"]
    assert next(gen) == ["putch", "o"]
    assert next(gen) == ["putch", "r"]
    assert next(gen) == ["putch", "l"]
    assert next(gen) == ["putch", "d"]

    with pytest.raises(StopIteration):
        next(gen)


def test_strip_ansi_codes():
    gen = ansi_color.strip_ansi_codes("hello \033[27mworld")
    assert gen == "hello world"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.interface.serial.i2c` class.
"""

import errno
import pytest
import smbus2
from unittest.mock import Mock, patch, call
from luma.core.interface.serial import i2c
import luma.core.error

from helpers import i2c_error, fib


smbus = Mock(unsafe=True)


def setup_function(function):
    smbus.reset_mock()


def test_init_device_not_found():
    port = 200
    address = 0x710
    path_name = f'/dev/i2c-{port}'
    fake_open = i2c_error(path_name, errno.ENOENT)

    with patch('os.open', fake_open):
        with pytest.raises(luma.core.error.DeviceNotFoundError) as ex:
            i2c(port=port, address=address)
        assert str(ex.value) == f'I2C device not found: {path_name}'


def test_init_device_permission_error():
    port = 1
    path_name = f'/dev/i2c-{port}'
    fake_open = i2c_error(path_name, errno.EACCES)

    with patch('os.open', fake_open):
        try:
            i2c(port=port)
        except luma.core.error.DevicePermissionError as ex:
            # permission error: device exists but no permission
            assert str(ex) == f'I2C device permission denied: {path_name}'


def test_init_device_address_error():
    address = 'foo'
    with pytest.raises(luma.core.error.DeviceAddressError) as ex:
        i2c(address=address)
    assert str(ex.value) == f'I2C device address invalid: {address}'


def test_init_no_bus():
    with patch.object(smbus2.SMBus, 'open') as mock:
        i2c(port=2, address=0x71)
    mock.assert_called_once_with(2)


def test_init_bus_provided():
    i2c(bus=smbus, address=0x71)
    smbus.open.assert_not_called()


def test_command():
    cmds = [3, 1, 4, 2]
    serial = i2c(bus=smbus, address=0x83)
    serial.command(*cmds)
    smbus.write_i2c_block_data.assert_called_once_with(0x83, 0x00, cmds)


def test_i2c_command_device_not_found_error():
    errorbus = Mock(unsafe=True)
    address = 0x71
    cmds = [3, 1, 4, 2]
    expected_error = OSError()

    try:
        for error_code in [errno.EREMOTEIO, errno.EIO]:
            expected_error.errno = error_code
            errorbus.write_i2c_block_data.side_effect = expected_error

            serial = i2c(bus=errorbus, address=address)
            with pytest.raises(luma.core.error.DeviceNotFoundError) as ex:
                serial.command(*cmds)

            assert str(ex.value) == f'I2C device not found on address: 0x{address:02X}'
    except AttributeError as e:
        # osx
        pytest.skip(str(e))


def test_i2c_data():
    data = list(fib(10))
    serial = i2c(bus=smbus, address=0x21)
    serial.data(data)
    smbus.write_i2c_block_data.assert_called_once_with(0x21, 0x40, data)


def test_i2c_data_chunked():
    data = list(fib(100))
    serial = i2c(bus=smbus, address=0x66)
    serial.data(data)
    calls = [call(0x66, 0x40, data[i:i + 32]) for i in range(0, 100, 32)]
    smbus.write_i2c_block_data.assert_has_calls(calls)


def test_cleanup():
    serial = i2c(bus=smbus, address=0x9F)
    serial._managed = True
    serial.cleanup()
    smbus.close.assert_called_once_with()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

from PIL import Image, ImageDraw
from luma.core.framebuffer import full_frame, diff_to_previous


im1 = Image.new("RGB", (40, 40))
draw = ImageDraw.Draw(im1)
draw.line((0, 0) + (39, 39), fill="white")

im2 = Image.new("RGB", (40, 40))
draw = ImageDraw.Draw(im2)
draw.line((0, 0) + (39, 39), fill="white")
draw.line((0, 39) + (39, 0), fill="white")


def test_full_frame():
    framebuffer = full_frame()
    redraws = list(framebuffer.redraw(im1))
    assert len(redraws) == 1
    assert redraws[0][0] == im1
    assert redraws[0][1] == im1.getbbox()


def test_diff_to_previous():
    framebuffer = diff_to_previous(num_segments=4)
    redraws = list(framebuffer.redraw(im1))

    # First redraw should be the full image
    assert len(redraws) == 1
    assert redraws[0][0] == im1
    assert redraws[0][1] == im1.getbbox()
    assert redraws[0][1] is not None

    # Redraw of same image should return empty changeset
    redraws = list(framebuffer.redraw(im1))
    assert len(redraws) == 0

    # Redraw of new image should return two changesets
    redraws = list(framebuffer.redraw(im2))
    assert len(redraws) == 2

    assert redraws[0][0] == im2.crop((0, 20, 20, 40))
    assert redraws[0][1] == (20, 0, 40, 20)

    assert redraws[1][0] == im2.crop((20, 0, 40, 20))
    assert redraws[1][1] == (0, 20, 20, 40)

    # Redraw of original image should return two changesets
    redraws = list(framebuffer.redraw(im1))
    assert len(redraws) == 2

    assert redraws[0][0] == im1.crop((20, 0, 40, 20))
    assert redraws[0][1] == (20, 0, 40, 20)

    assert redraws[1][0] == im1.crop((0, 20, 20, 40))
    assert redraws[1][1] == (0, 20, 20, 40)


def test_diff_to_previous_debug():
    framebuffer = diff_to_previous(num_segments=4, debug=True)
    redraws = list(framebuffer.redraw(im1))

    # First redraw should be the full image unchanged
    assert len(redraws) == 1
    assert redraws[0][0] == im1

    # Redraw of new image should return two changesets with red borders around them
    redraws = list(framebuffer.redraw(im2))
    assert len(redraws) == 2

    first_changeset_image = im2.copy().crop((20, 0, 40, 20))
    draw = ImageDraw.Draw(first_changeset_image)
    draw.rectangle((0, 0, 19, 19), outline="red")
    del draw
    assert redraws[0][0] == first_changeset_image

    second_changeset_image = im2.copy().crop((0, 20, 20, 40))
    draw = ImageDraw.Draw(second_changeset_image)
    draw.rectangle((0, 0, 19, 19), outline="red")
    del draw
    assert redraws[1][0] == second_changeset_image



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2020 Richard Hull and contributors
# See LICENSE.rst for details.


import pytest

from luma.core.render import canvas
from luma.core.device import dummy
from luma.core.virtual import history

import baseline_data


def test_restore_throws_error_when_empty():
    device = dummy()
    hist = history(device)
    assert len(hist) == 0
    with pytest.raises(IndexError):
        hist.restore()


def test_save_and_restore_reverts_image():
    device = dummy()
    hist = history(device)

    with canvas(hist) as draw:
        baseline_data.primitives(hist, draw)

    img1 = device.image
    hist.savepoint()
    assert len(hist) == 1

    with canvas(hist) as draw:
        draw.text((10, 10), text="Hello", fill="white")

    img2 = device.image
    assert img1 != img2
    hist.restore()
    img3 = device.image
    assert img1 == img3
    assert len(hist) == 0


def test_drop_and_restore():
    device = dummy()
    hist = history(device)

    copies = []
    for i in range(10):
        with canvas(hist) as draw:
            draw.text((10, 10), text=f"Hello {i}", fill="white")
        hist.savepoint()
        copies.append(device.image)

    hist.restore()
    assert device.image == copies[9]
    hist.restore(drop=2)
    assert device.image == copies[6]
    hist.restore(drop=4)
    assert device.image == copies[1]
    assert len(hist) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Richard Hull and contributors
# See LICENSE.rst for details.

from pathlib import Path

from PIL import Image, ImageFont
from luma.core.device import dummy
from luma.core.virtual import character

from helpers import get_reference_file, get_reference_image, assert_identical_image


def test_init():
    path = get_reference_file(Path('font').joinpath('hd44780a02.pil'))
    fnt = ImageFont.load(path)
    device = dummy(width=80, height=16, mode="1")
    character(device, font=fnt)
    assert device.image == Image.new("1", (80, 16))


def test_setter_getter():
    fnt_path = get_reference_file(Path('font').joinpath('hd44780a02.pil'))
    img_path = get_reference_image('character_golden_ratio.png')

    with open(img_path, 'rb') as img:
        fnt = ImageFont.load(fnt_path)
        reference = Image.open(img)
        device = dummy(width=80, height=16, mode="1")
        c = character(device, font=fnt)
        c.text = "1.61803398875\n1.61803398875"
        assert str(c.text) == "1.61803398875\n1.61803398875"

        assert_identical_image(reference, device.image, img_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.


import pytest

from luma.core.mixin import capabilities


def test_display_not_implemented():
    cap = capabilities()
    with pytest.raises(NotImplementedError):
        cap.display('foo')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.sprite_system.spritesheet` class.
"""

import pytest
import PIL

from luma.core.sprite_system import spritesheet


data = {
    'image': "tests/reference/images/runner.png",
    'frames': {
        'width': 64,
        'height': 67,
        'regX': 0,
        'regY': 2
    },
    'animations': {
        'run-right': {
            'frames': range(19, 9, -1),
            'next': "run-right",
        },
        'run-left': {
            'frames': range(0, 10)
        },
        'run-slow': {
            'frames': range(0, 10),
            'speed': 0.5
        },
        'composite': {
            'frames': [1, 2, 3, 'run-left', 4, 5, 6, 'run-left']
        }
    }
}


def test_init():
    sheet = spritesheet(**data)
    # Reframed by 2px
    assert sheet.width == 640
    assert sheet.height == 134
    assert sheet.frames.count == 20
    assert sheet.frames.width == 64
    assert sheet.frames.height == 67
    assert sheet.frames.size == (64, 67)
    assert sheet.cache == {}


def test_len():
    sheet = spritesheet(**data)
    assert len(sheet) == sheet.frames.count


def test_caching():
    sheet = spritesheet(**data)
    assert sheet.cache == {}
    frame = sheet[17]
    assert list(sheet.cache.keys()) == [17]
    assert sheet.cache[17] == frame


def test_get():
    w = 64
    h = 67
    sheet = spritesheet(**data)
    frame = sheet[16]
    expected = sheet.image.crop((w * 6, h * 1, w * 7, h * 2))
    assert frame == expected


def test_get_string():
    sheet = spritesheet(**data)
    with pytest.raises(TypeError):
        sheet['banana']


def test_get_outofrange():
    sheet = spritesheet(**data)
    with pytest.raises(IndexError):
        sheet[-2]
    with pytest.raises(IndexError):
        sheet[3002]


def test_animate_unknown_seq():
    sheet = spritesheet(**data)
    with pytest.raises(AttributeError):
        list(sheet.animate("unknown"))


def test_animate_finite_seq():
    sheet = spritesheet(**data)
    frames = list(sheet.animate("run-left"))
    assert len(frames) == 10
    for i, frame in enumerate(frames):
        assert isinstance(frame, PIL.Image.Image)
        assert sheet[i] == frame


def test_animate_slow_seq():
    sheet = spritesheet(**data)
    frames = list(sheet.animate("run-slow"))
    assert len(frames) == 20
    for i, frame in enumerate(frames):
        assert isinstance(frame, PIL.Image.Image)
        assert sheet[i // 2] == frame


def test_animate_infinite_seq():
    sheet = spritesheet(**data)
    runner = sheet.animate("run-right")
    frames = []
    for _ in range(50):
        frames.append(next(runner))

    expected = list(range(19, 9, -1)) * 5

    for i, frame_ref in enumerate(expected):
        assert isinstance(frames[i], PIL.Image.Image)
        assert sheet[frame_ref] == frames[i]


def test_animate_subroutine():
    sheet = spritesheet(**data)
    frames = list(sheet.animate("composite"))

    expected = [1, 2, 3, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 4, 5, 6, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    assert len(frames) == 26
    for i, frame_ref in enumerate(expected):
        assert isinstance(frames[i], PIL.Image.Image)
        assert sheet[frame_ref] == frames[i]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.sprite_system.framerate_regulator` class.
"""

from luma.core.util import perf_counter
from luma.core.sprite_system import framerate_regulator


def test_init_default():
    regulator = framerate_regulator()
    assert regulator.start_time is None
    assert regulator.last_time is None
    assert regulator.called == 0
    before = perf_counter()
    with regulator:
        pass
    after = perf_counter()

    assert regulator.max_sleep_time == 1 / 16.67
    assert before <= regulator.start_time <= after
    assert regulator.called == 1


def test_init_unlimited():
    regulator = framerate_regulator(fps=0)
    before = perf_counter()
    with regulator:
        pass
    after = perf_counter()

    assert regulator.max_sleep_time == -1
    assert before <= regulator.start_time <= after
    assert regulator.called == 1


def test_init_30fps():
    regulator = framerate_regulator(fps=30)
    before = perf_counter()
    with regulator:
        pass
    after = perf_counter()

    assert regulator.max_sleep_time == 1 / 30.00
    assert before <= regulator.start_time <= after
    assert regulator.called == 1


def test_sleep():
    regulator = framerate_regulator(fps=100.00)
    before = perf_counter()
    for _ in range(200):
        with regulator:
            pass
    after = perf_counter()

    assert regulator.called == 200
    assert after - before >= 2.0


def test_effective_FPS():
    regulator = framerate_regulator(fps=30)
    assert regulator.effective_FPS() == 0


def test_average_transit_time():
    regulator = framerate_regulator(fps=30)
    with regulator:
        pass
    assert regulator.average_transit_time() > 0



# -*- coding: utf-8 -*-
# Copyright (c) 2017-2022 Richard Hull and contributors
# See LICENSE.rst for details.

import sys
from time import perf_counter_ns


if sys.version_info.major == 3:
    unicode = str


class mutable_string(object):

    def __init__(self, value):
        assert isinstance(value, str) or isinstance(value, unicode)
        self.target = value

    def __getattr__(self, attr):
        return self.target.__getattribute__(attr)

    def __getitem__(self, key):
        return self.target[key]

    def __setitem__(self, key, value):
        assert isinstance(value, str) or isinstance(value, unicode)
        tmp = list(self.target)
        tmp[key] = value
        self.target = ("" if isinstance(self.target, str) else u"").join(tmp)

    def __delitem__(self, key):
        tmp = list(self.target)
        del tmp[key]
        self.target = "".join(tmp)

    def __len__(self):
        return len(self.target)

    def __iter__(self):
        return iter(self.target)

    def __str__(self):
        return self.target

    def __repr__(self):
        return repr(self.target)

    def __eq__(self, other):
        if isinstance(self.target, unicode):
            return self.target == unicode(other)
        else:
            return self.target == str(other)

    def __hash__(self):
        return hash(self.target)


class observable(object):
    """
    Wraps any container object such that on inserting, updating or deleting,
    an observer is notified with a payload of the target. All other special
    name methods are passed through parameters unhindered.
    """
    def __init__(self, target, observer):
        self.target = target
        self.observer = observer
        self.observer(self.target)

    def __getattr__(self, attr):
        return self.target.__getattribute__(attr)

    def __getitem__(self, key):
        return self.target.__getitem__(key)

    def __setitem__(self, key, value):
        self.target.__setitem__(key, value)
        self.observer(self.target)

    def __delitem__(self, key):
        self.target.__delitem__(key)
        self.observer(self.target)

    def __len__(self):
        return self.target.__len__()

    def __iter__(self):
        return self.target.__iter__()

    def __str__(self):
        return self.target.__str__()

    def __repr__(self):
        return self.target.__repr__()


def from_16_to_8(data):
    """
    Utility function to take a list of 16 bit values and turn it into
    a list of 8 bit values

    :param data: list of 16 bit values to convert
    :type data: list
    :return: a list of 8 bit values
    :rtype: list

    .. versionadded:: 1.16.0
    """
    return [f(x) for x in data for f in (lambda x: (x & 0xFF00) >> 8, lambda x: 0xFF & x)]


def from_8_to_16(data):
    """
    Utility function to take a list of 8 bit values and turn it into a list
    of signed 16 bit integers

    :param data: list of 8 bit values to convert
    :type data: list
    :return: a list of 16 bit values
    :rtype: list

    .. versionadded:: 1.16.0
    """
    return [unsigned_16_to_signed(((data[i] & 0xFF) << 8) + (data[i + 1] & 0xFF))
        for i in range(0, len(data), 2)] if data is not None else None


def unsigned_16_to_signed(value):
    """
    Utility function to convert unsigned 16 bit value to a signed value

    :param value: the 16 bit value that needs to be converted
    :type value: int
    :return: a signed integer
    :rtype: int

    .. versionadded:: 1.16.0
    """
    return ((value) & 0x7FFF) - (0x8000 & (value))


def bytes_to_nibbles(data):
    """
    Utility function to take a list of bytes (8 bit values) and turn it into
    a list of nibbles (4 bit values)

    :param data: a list of 8 bit values that will be converted
    :type data: list
    :return: a list of 4 bit values
    :rtype: list

    .. versionadded:: 1.16.0
    """
    return [f(x) for x in data for f in (lambda x: x >> 4, lambda x: 0x0F & x)]


def perf_counter():
    """
    :py:func:`time.perf_counter_ns` in seconds

    :return: the value (in fractional seconds) of a performance counter
    :rtype: float

    .. versionadded:: 2.4.0
    """
    return perf_counter_ns() / 1e9



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.virtual.viewport` class and associated
helpers.
"""

import time

from PIL import Image

from luma.core.device import dummy
from luma.core.render import canvas
from luma.core.virtual import range_overlap, hotspot, snapshot, viewport

import baseline_data
from helpers import get_reference_image, assert_identical_image


def overlap(box1, box2):
    l1, t1, r1, b1 = box1
    l2, t2, r2, b2 = box2
    return range_overlap(l1, r1, l2, r2) and range_overlap(t1, b1, t2, b2)


box1 = [0, 0, 64, 64]
box2 = [64, 0, 128, 64]
box3 = [128, 0, 192, 64]
box4 = [192, 0, 256, 64]


def test_range_overlap_over12():
    viewport = [0, 0, 128, 64]
    assert overlap(viewport, box1) is True
    assert overlap(viewport, box2) is True
    assert overlap(viewport, box3) is False
    assert overlap(viewport, box4) is False


def test_range_overlap_over123():
    viewport = [30, 0, 158, 64]
    assert overlap(viewport, box1) is True
    assert overlap(viewport, box2) is True
    assert overlap(viewport, box3) is True
    assert overlap(viewport, box4) is False


def test_range_overlap_over23():
    viewport = [64, 0, 192, 64]
    assert overlap(viewport, box1) is False
    assert overlap(viewport, box2) is True
    assert overlap(viewport, box3) is True
    assert overlap(viewport, box4) is False


def test_range_overlap_over234():
    viewport = [100, 0, 228, 64]
    assert overlap(viewport, box1) is False
    assert overlap(viewport, box2) is True
    assert overlap(viewport, box3) is True
    assert overlap(viewport, box4) is True


def test_range_overlap_over34():
    viewport = [128, 0, 256, 64]
    assert overlap(viewport, box1) is False
    assert overlap(viewport, box2) is False
    assert overlap(viewport, box3) is True
    assert overlap(viewport, box4) is True


def test_range_overlap_over4():
    viewport = [192, 0, 256, 64]
    assert overlap(viewport, box1) is False
    assert overlap(viewport, box2) is False
    assert overlap(viewport, box3) is False
    assert overlap(viewport, box4) is True


def test_range_overlap_over_none():
    viewport = [256, 0, 384, 64]
    assert overlap(viewport, box1) is False
    assert overlap(viewport, box2) is False
    assert overlap(viewport, box3) is False
    assert overlap(viewport, box4) is False


def test_snapshot_last_updated():
    interval = 0.5

    def draw_fn(draw, width, height):
        assert height == 10
        assert width == 10

    sshot = snapshot(10, 10, draw_fn, interval)
    assert sshot.last_updated == -interval
    assert sshot.should_redraw() is True
    sshot.paste_into(Image.new("RGB", (10, 10)), (0, 0))
    assert sshot.should_redraw() is False
    time.sleep(interval * 1.5)
    assert sshot.should_redraw() is True
    sshot.paste_into(Image.new("RGB", (10, 10)), (0, 0))
    assert sshot.last_updated > 0.0
    assert sshot.should_redraw() is False


def test_viewport_set_position():
    img_path = get_reference_image('set_position.png')

    with open(img_path, 'rb') as p:
        reference = Image.open(p)
        device = dummy()
        virtual = viewport(device, 200, 200)

        # Use the same drawing primitives as the demo
        with canvas(virtual) as draw:
            baseline_data.primitives(virtual, draw)

        virtual.set_position((20, 30))

        assert_identical_image(reference, device.image, img_path)


def test_viewport_hotspot():
    img_path = get_reference_image('hotspot.png')

    with open(img_path, 'rb') as p:
        reference = Image.open(p)
        device = dummy()
        virtual = viewport(device, 200, 200)

        def draw_fn(draw, width, height):
            baseline_data.primitives(device, draw)

        widget = hotspot(device.width, device.height, draw_fn)

        virtual.add_hotspot(widget, (19, 56))
        virtual.set_position((28, 30))
        virtual.remove_hotspot(widget, (19, 56))

        assert_identical_image(reference, device.image, img_path)


def test_viewport_dithering():
    img_path = get_reference_image('hotspot_dither.png')

    with open(img_path, 'rb') as p:
        reference = Image.open(p)
        device = dummy(mode="1")
        virtual = viewport(device, 200, 200, mode='RGBA', dither=True)

        def draw_fn(draw, width, height):
            draw.rectangle((0, 0, width / 2, height / 2), fill="red")
            draw.rectangle((width / 2, 0, width, height / 2), fill="yellow")
            draw.rectangle((0, height / 2, width / 2, height), fill="blue")
            draw.rectangle((width / 2, height / 2, width, height), fill="white")

        widget = hotspot(device.width, device.height, draw_fn)

        virtual.add_hotspot(widget, (19, 56))
        virtual.set_position((28, 30))
        virtual.remove_hotspot(widget, (19, 56))

        assert_identical_image(reference, device.image, img_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-20 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.interface.serial.gpio_cs_spi` class.
"""

import pytest
from unittest.mock import Mock, call

from luma.core.interface.serial import gpio_cs_spi
import luma.core.error

from helpers import get_spidev, rpi_gpio_missing, fib, assert_only_cleans_whats_setup


spidev = Mock(unsafe=True)
gpio = Mock(unsafe=True)


def setup_function(function):
    spidev.reset_mock()
    gpio.reset_mock()
    gpio.BCM = 1
    gpio.RST = 2
    gpio.DC = 3
    gpio.OUT = 4
    gpio.HIGH = 5
    gpio.LOW = 6


def verify_gpio_cs_spi_init(port, device, bus_speed_hz=8000000, dc=24, rst=25, cs=23):
    spidev.open.assert_called_once_with(port, device)
    assert spidev.max_speed_hz == bus_speed_hz
    gpio.setmode.assert_not_called()
    gpio.setup.assert_has_calls([call(dc, gpio.OUT), call(rst, gpio.OUT), call(cs, gpio.OUT, initial=gpio.HIGH)])


def test_init():
    port = 5
    device = 2
    bus_speed = 16000000
    dc = 17
    rst = 11
    cs = 23

    gpio_cs_spi(gpio=gpio, spi=spidev, port=port, device=device, bus_speed_hz=16000000,
        gpio_DC=dc, gpio_RST=rst, gpio_CS=cs)
    verify_gpio_cs_spi_init(port, device, bus_speed, dc, rst, cs)
    gpio.output.assert_has_calls([
        call(rst, gpio.LOW),
        call(rst, gpio.HIGH)
    ])


def test_init_invalid_bus_speed():
    with pytest.raises(AssertionError):
        gpio_cs_spi(gpio=gpio, spi=spidev, port=5, device=2, bus_speed_hz=942312, gpio_CS=23)


def test_command():
    cmds = [3, 1, 4, 2]
    serial = gpio_cs_spi(gpio=gpio, spi=spidev, port=9, device=1, gpio_CS=23)
    serial.command(*cmds)
    verify_gpio_cs_spi_init(9, 1)
    gpio.output.assert_has_calls([call(25, gpio.HIGH), call(24, gpio.LOW), call(23, gpio.LOW), call(23, gpio.HIGH)])
    spidev.writebytes.assert_called_once_with(cmds)


def test_data():
    data = list(fib(100))
    serial = gpio_cs_spi(gpio=gpio, spi=spidev, port=9, device=1, gpio_CS=23)
    serial.data(data)
    verify_gpio_cs_spi_init(9, 1)
    gpio.output.assert_has_calls([call(25, gpio.HIGH), call(24, gpio.HIGH), call(23, gpio.LOW), call(23, gpio.HIGH)])
    spidev.writebytes.assert_called_once_with(data)


def test_cleanup():
    serial = gpio_cs_spi(gpio=gpio, spi=spidev, port=9, device=1, gpio_CS=23)
    serial._managed = True
    serial.cleanup()
    verify_gpio_cs_spi_init(9, 1)
    spidev.close.assert_called_once_with()
    assert_only_cleans_whats_setup(gpio)


def test_init_device_not_found():
    spidev = get_spidev()
    port = 1234
    with pytest.raises(luma.core.error.DeviceNotFoundError) as ex:
        gpio_cs_spi(gpio=gpio, spi=spidev.SpiDev(), port=port, gpio_CS=23)
    assert str(ex.value) == 'SPI device not found'


def test_unsupported_gpio_platform():
    try:
        gpio_cs_spi(spi=spidev, port=9, device=1, gpio_CS=23)
    except luma.core.error.UnsupportedPlatform as ex:
        assert str(ex) == 'GPIO access not available'
    except ImportError:
        pytest.skip(rpi_gpio_missing)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2021 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:mod:`luma.core.cmdline` module.
"""

import pytest
import errno
from unittest.mock import patch, Mock

from luma.core import cmdline, error
from luma.core.interface.serial import __all__ as serial_iface_types
from luma.core.interface.parallel import __all__ as parallel_iface_types

from helpers import (get_reference_file, i2c_error, rpi_gpio_missing,
    spidev_missing, skip_unsupported_platform)


test_config_file = get_reference_file('config-test.txt')


class test_spi_opts(object):
    spi_port = 0
    spi_mode = 0
    spi_device = 0
    spi_bus_speed = 8000000
    spi_transfer_size = 4096
    gpio_data_command = 24
    gpio_reset = 25
    gpio_backlight = 18
    gpio_reset_hold_time = 0
    gpio_reset_release_time = 0

    interface = 'spi'
    framebuffer = 'diff_to_previous'
    num_segments = 25
    debug = False
    framebuffer_device = None


def test_get_interface_types():
    """
    Enumerate interface types.
    """
    assert cmdline.get_interface_types() == serial_iface_types + parallel_iface_types


def test_ensure_cmdline_opt_contains_all_interfaces():
    """
    Checks that the cmdline make_interface factory contains initializers for all interface classes
    """
    class opts:
        pass

    factory = cmdline.make_interface(opts)
    for interface in cmdline.get_interface_types():
        assert hasattr(factory, interface)


def test_get_display_types():
    """
    Enumerate display types.
    """
    assert list(cmdline.get_display_types().keys()) == \
        cmdline.get_supported_libraries()


def test_get_choices_unknown_module():
    """
    :py:func:`luma.core.cmdline.get_choices` returns an empty list when
    trying to inspect an unknown module.
    """
    result = cmdline.get_choices('foo')
    assert result == []


def test_get_library_version():
    """
    :py:func:`luma.core.cmdline.get_library_version` returns the version number
    for the specified library name.
    """
    lib_name = 'hotscreenz'
    lib_version = '0.1.2'

    # set version nr for fake luma library
    luma_fake_lib = Mock()
    luma_fake_lib.__version__ = lib_version

    # version is found
    with patch.dict('sys.modules', {'luma.' + lib_name: luma_fake_lib}):
        assert cmdline.get_library_version(lib_name) == lib_version

    # no version for module without __version__ attribute
    lib_name = 'no_version'
    luma_without_version_lib = Mock()
    with patch.dict('sys.modules', {'luma.' + lib_name: luma_without_version_lib}):
        assert cmdline.get_library_version(lib_name) is None

    # no version for non-existing module
    assert cmdline.get_library_version('foo') is None


def test_get_library_for_display_type():
    """
    :py:func:`luma.core.cmdline.get_library_for_display_type` returns the
    the library name for a particular display.
    """
    display_type = 'coolscreen'
    lib_name = 'screens'

    with patch('luma.core.cmdline.get_display_types') as mocka:
        mocka.return_value = {
            lib_name: [display_type, 'bar'],
            'emulator': ['x', 'y']
        }
        assert cmdline.get_library_for_display_type(display_type) == lib_name


def test_load_config_file_parse():
    """
    :py:func:`luma.core.cmdline.load_config` parses a text file and returns a
    list of arguments.
    """
    result = cmdline.load_config(test_config_file)
    assert result == [
        '--display=capture',
        '--width=800',
        '--height=8600',
        '--spi-bus-speed=16000000'
    ]


def test_create_parser():
    """
    :py:func:`luma.core.cmdline.create_parser` returns an argument parser
    instance.
    """
    with patch.dict('sys.modules', **{
            'luma.emulator': Mock(),
            'luma.emulator.render': Mock(),
        }):
        with patch('luma.core.cmdline.get_display_types') as mocka:
            mocka.return_value = {
                'foo': ['a', 'b'],
                'bar': ['c', 'd'],
                'emulator': ['e', 'f']
            }
            parser = cmdline.create_parser(description='test')
            args = parser.parse_args(['-f', test_config_file])
            assert args.config == test_config_file


def test_make_interface_noop():
    """
    :py:func:`luma.core.cmdline.make_interface.noop` returns an ``noop` instance.
    """
    class opts:
        interface = 'noop'

    factory = cmdline.make_interface(opts)
    assert 'luma.core.interface.serial.noop' in repr(factory.noop())


def test_make_interface_i2c():
    """
    :py:func:`luma.core.cmdline.make_interface.i2c` returns an I2C instance.
    """
    class opts:
        i2c_port = 200
        i2c_address = 0x710

    path_name = f'/dev/i2c-{opts.i2c_port}'
    fake_open = i2c_error(path_name, errno.ENOENT)
    factory = cmdline.make_interface(opts)

    with patch('os.open', fake_open):
        with pytest.raises(error.DeviceNotFoundError):
            factory.i2c()


def test_make_interface_spi():
    """
    :py:func:`luma.core.cmdline.make_interface.spi` returns an SPI instance.
    """
    try:
        factory = cmdline.make_interface(test_spi_opts)
        assert 'luma.core.interface.serial.spi' in repr(factory.spi())
    except ImportError:
        # non-rpi platform, e.g. macos
        pytest.skip(rpi_gpio_missing)
    except error.UnsupportedPlatform as e:
        # non-rpi platform, e.g. ubuntu 64-bit
        skip_unsupported_platform(e)


def test_make_interface_gpio_cs_spi():
    """
    :py:func:`luma.core.cmdline.make_interface.gpio_cs_spi` returns an gpio_cs_spi instance.
    """
    class opts(test_spi_opts):
        interface = 'gpio_cs_spi'
        spi_cs_high = True
        gpio_chip_select = 4

    try:
        factory = cmdline.make_interface(opts)
        assert 'luma.core.interface.serial.spi' in repr(factory.gpio_cs_spi())
    except ImportError:
        # non-rpi platform, e.g. macos
        pytest.skip(rpi_gpio_missing)
    except error.UnsupportedPlatform as e:
        # non-rpi platform, e.g. ubuntu 64-bit
        skip_unsupported_platform(e)


def test_make_interface_spi_alt_gpio():
    """
    :py:func:`luma.core.cmdline.make_interface.spi` returns an SPI instance
    when using an alternative GPIO implementation.
    """
    class opts(test_spi_opts):
        gpio = 'fake_gpio'

    with patch.dict('sys.modules', **{
            'fake_gpio': Mock(unsafe=True)
        }):
        try:
            factory = cmdline.make_interface(opts)
            assert 'luma.core.interface.serial.spi' in repr(factory.spi())
        except ImportError:
            pytest.skip(spidev_missing)
        except error.DeviceNotFoundError as e:
            # non-rpi platform, e.g. ubuntu 64-bit
            skip_unsupported_platform(e)


def test_make_interface_bitbang():
    """
    :py:func:`luma.core.cmdline.make_interface.bitbang` returns an BitBang instance.
    """
    try:
        factory = cmdline.make_interface(test_spi_opts)
        assert 'luma.core.interface.serial.bitbang' in repr(factory.bitbang())
    except ImportError:
        # non-rpi platform, e.g. macos
        pytest.skip(rpi_gpio_missing)
    except error.UnsupportedPlatform as e:
        # non-rpi platform, e.g. ubuntu 64-bit
        skip_unsupported_platform(e)


def test_make_interface_pcf8574():
    """
    :py:func:`luma.core.cmdline.make_interface.pcf8574` returns an pcf8574 instance.
    """
    class opts:
        i2c_port = 200
        i2c_address = 0x710

    path_name = f'/dev/i2c-{opts.i2c_port}'
    fake_open = i2c_error(path_name, errno.ENOENT)
    factory = cmdline.make_interface(opts)

    with patch('os.open', fake_open):
        with pytest.raises(error.DeviceNotFoundError):
            factory.pcf8574()


def test_make_interface_bitbang_6800():
    """
    :py:func:`luma.core.cmdline.make_interface.bitbang_6800` returns a Bitbang-6800 instance.
    """
    class opts:
        pass

    try:
        factory = cmdline.make_interface(opts)
        assert 'luma.core.interface.parallel.bitbang_6800' in repr(factory.bitbang_6800())
    except ImportError:
        # non-rpi platform, e.g. macos
        pytest.skip(rpi_gpio_missing)
    except error.UnsupportedPlatform as e:
        # non-rpi platform, e.g. ubuntu 64-bit
        skip_unsupported_platform(e)


def test_make_interface_bitbang_6800_alt_gpio():
    """
    :py:func:`luma.core.cmdline.make_interface.bitbang_6800` returns an Bitbang-6800 instance
    when using an alternative GPIO implementation.
    """
    class opts():
        gpio = 'fake_gpio'

    with patch.dict('sys.modules', **{
            'fake_gpio': Mock(unsafe=True)
        }):
        try:
            factory = cmdline.make_interface(opts)
            assert 'luma.core.interface.parallel.bitbang_6800' in repr(factory.bitbang_6800())
        except ImportError:
            pytest.skip(spidev_missing)
        except error.DeviceNotFoundError as e:
            # non-rpi platform, e.g. ubuntu 64-bit
            skip_unsupported_platform(e)


def test_create_device():
    """
    :py:func:`luma.core.cmdline.create_device` returns ``None`` for unknown
    displays.
    """
    class args:
        display = 'foo'
    assert cmdline.create_device(args) is None


def test_create_device_oled():
    """
    :py:func:`luma.core.cmdline.create_device` supports OLED displays.
    """
    display_name = 'oled1234'
    display_types = {'oled': [display_name]}

    class args(test_spi_opts):
        display = display_name

    module_mock = Mock()
    module_mock.oled.device.oled1234.return_value = display_name
    with patch.dict('sys.modules', **{
            # mock luma.oled package
            'luma': module_mock,
            'luma.oled': module_mock,
            'luma.oled.device': module_mock
        }):
        try:
            device = cmdline.create_device(args, display_types=display_types)
            assert device == display_name
        except ImportError:
            pytest.skip(rpi_gpio_missing)
        except error.UnsupportedPlatform as e:
            # non-rpi platform
            skip_unsupported_platform(e)


def test_create_device_lcd():
    """
    :py:func:`luma.core.cmdline.create_device` supports LCD displays.
    """
    display_name = 'lcd1234'
    display_types = {'lcd': [display_name]}

    class args(test_spi_opts):
        display = display_name
        gpio = 'fake_gpio'
        backlight_active = 'low'

    module_mock = Mock()
    module_mock.lcd.device.lcd1234.return_value = display_name
    with patch.dict('sys.modules', **{
            # mock spidev and luma.lcd packages
            'fake_gpio': module_mock,
            'spidev': module_mock,
            'luma': module_mock,
            'luma.lcd': module_mock,
            'luma.lcd.aux': module_mock,
            'luma.lcd.device': module_mock
        }):
        device = cmdline.create_device(args, display_types=display_types)
        assert device == display_name


def test_create_device_led_matrix():
    """
    :py:func:`luma.core.cmdline.create_device` supports LED matrix displays.
    """
    display_name = 'matrix1234'
    display_types = {'led_matrix': [display_name]}

    class args(test_spi_opts):
        display = display_name

    module_mock = Mock()
    module_mock.led_matrix.device.matrix1234.return_value = display_name
    with patch.dict('sys.modules', **{
            # mock spidev and luma.led_matrix packages
            'spidev': module_mock,
            'luma': module_mock,
            'luma.led_matrix': module_mock,
            'luma.led_matrix.device': module_mock
        }):
        device = cmdline.create_device(args, display_types=display_types)
        assert device == display_name


def test_create_device_emulator():
    """
    :py:func:`luma.core.cmdline.create_device` supports emulators.
    """
    display_name = 'emulator1234'
    display_types = {'emulator': [display_name]}

    class args(test_spi_opts):
        display = display_name

    module_mock = Mock()
    module_mock.emulator.device.emulator1234.return_value = display_name
    with patch.dict('sys.modules', **{
            # mock spidev and luma.emulator packages
            'spidev': module_mock,
            'luma': module_mock,
            'luma.emulator': module_mock,
            'luma.emulator.device': module_mock
        }):
        device = cmdline.create_device(args, display_types=display_types)
        assert device == display_name


def test_create_device_core():
    """
    :py:func:`luma.core.cmdline.create_device` supports code devices.
    """
    display_name = 'coredevice1234'
    display_types = {'core': [display_name]}

    class args(test_spi_opts):
        display = display_name

    module_mock = Mock()
    module_mock.core.device.coredevice1234.return_value = display_name
    with patch.dict('sys.modules', **{
            # mock luma.core package
            'luma': module_mock,
            'luma.core': module_mock,
            'luma.core.device': module_mock
        }):
        device = cmdline.create_device(args, display_types=display_types)
        assert device == display_name


@patch('pyftdi.spi.SpiController')
def test_make_interface_ftdi_spi(mock_controller):
    """
    :py:func:`luma.core.cmdline.make_interface.ftdi_spi` returns an SPI instance.
    """
    class opts(test_spi_opts):
        ftdi_device = 'ftdi://dummy'
        gpio_data_command = 5
        gpio_reset = 6
        gpio_backlight = 7

    factory = cmdline.make_interface(opts)
    assert 'luma.core.interface.serial.spi' in repr(factory.ftdi_spi())


@patch('pyftdi.i2c.I2cController')
def test_make_interface_ftdi_i2c(mock_controller):
    """
    :py:func:`luma.core.cmdline.make_interface.ftdi_i2c` returns an I2C instance.
    """
    class opts:
        ftdi_device = 'ftdi://dummy'
        i2c_port = 200
        i2c_address = 0x710

    factory = cmdline.make_interface(opts)
    assert 'luma.core.interface.serial.i2c' in repr(factory.ftdi_i2c())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2020 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.legacy.fonts` class.
"""

from PIL import Image
from luma.core.device import dummy
from luma.core.render import canvas
from luma.core.legacy import text, textsize
import luma.core.legacy.font
from luma.core.legacy.font import proportional

import pytest
from helpers import get_reference_image, assert_identical_image


charset = ''.join(chr(i) for i in range(256))


@pytest.mark.parametrize("fontname",
                         ["ATARI_FONT", "CP437_FONT", "SINCLAIR_FONT",
                          "SPECCY_FONT", "LCD_FONT", "UKR_FONT", "TINY_FONT",
                          "SEG7_FONT"])
def test_font(fontname):
    font = getattr(luma.core.legacy.font, fontname)
    w, h = textsize(charset, proportional(font))
    device = dummy(width=w, height=h, mode="1")
    with canvas(device) as draw:
        text(draw, (0, 0), charset, "white", font=proportional(font))

    img_path = get_reference_image(f'{fontname}.png')
    with open(img_path, 'rb') as fp:
        reference = Image.open(fp)
        assert_identical_image(reference, device.image, img_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.


import pytest

from luma.core.legacy.font import proportional, CP437_FONT


def test_narrow_char():
    font = proportional(CP437_FONT)
    assert font[ord('!')] == [6, 95, 95, 6, 0]


def test_wide_char():
    font = proportional(CP437_FONT)
    assert font[ord('W')] == CP437_FONT[ord('W')]


def test_space_char():
    font = proportional(CP437_FONT)
    assert font[ord(' ')] == [0] * 4


def test_doublequote_char():
    font = proportional(CP437_FONT)
    assert font[ord('"')] == [7, 7, 0, 7, 7, 0]


def test_trim_not_nonzero():
    font = proportional(CP437_FONT)
    assert font._trim([0, 0, 0, 0]) == []


def test_unicode_not_supported():
    font = proportional(CP437_FONT)
    with pytest.raises(IndexError) as ex:
        font[ord("😀")]
    assert str(ex.value) == 'Font does not have ASCII code: 128512'



# -*- coding: utf-8 -*-
# Copyright (c) 2017-2022 Richard Hull and contributors
# See LICENSE.rst for details.

from time import sleep
from textwrap import TextWrapper

from PIL import Image, ImageDraw, ImageFont

from luma.core import mixin, ansi_color
from luma.core.threadpool import threadpool
from luma.core.render import canvas
from luma.core.util import mutable_string, observable, perf_counter


pool = threadpool(4)


def calc_bounds(xy, entity):
    """
    For an entity with width and height attributes, determine
    the bounding box if were positioned at ``(x, y)``.
    """
    left, top = xy
    right, bottom = left + entity.width, top + entity.height
    return [left, top, right, bottom]


def range_overlap(a_min, a_max, b_min, b_max):
    """
    Neither range is completely greater than the other.
    """
    return (a_min < b_max) and (b_min < a_max)


class viewport(mixin.capabilities):
    """
    The viewport offers a positionable window into a larger resolution pseudo-display,
    that also supports the concept of hotspots (which act like live displays).

    :param device: The device to project the enlarged pseudo-display viewport onto.
    :param width: The number of horizontal pixels.
    :type width: int
    :param height: The number of vertical pixels.
    :type height: int
    :param mode: The supported color model, one of ``"1"``, ``"RGB"`` or
        ``"RGBA"`` only.
    :type mode: str
    :param dither: By default, any color (other than black) will be `generally`
        treated as white when displayed on monochrome devices. However, this behaviour
        can be changed by adding ``dither=True`` and the image will be converted from RGB
        space into a 1-bit monochrome image where dithering is employed to differentiate
        colors at the expense of resolution.
    :type dither: bool
    """
    def __init__(self, device, width, height, mode=None, dither=False):
        self.capabilities(width, height, rotate=0, mode=mode or device.mode)
        if hasattr(device, "segment_mapper"):
            self.segment_mapper = device.segment_mapper
        self._device = device
        self._backing_image = Image.new(self.mode, self.size)
        self._position = (0, 0)
        self._hotspots = []
        self._dither = dither

    def display(self, image):
        assert image.mode == self.mode
        assert image.size == self.size

        self._backing_image.paste(image)
        self.refresh()

    def set_position(self, xy):
        self._position = xy
        self.refresh()

    def add_hotspot(self, hotspot, xy):
        """
        Add the hotspot at ``(x, y)``. The hotspot must fit inside the bounds
        of the virtual device. If it does not then an ``AssertError`` is
        raised.
        """
        (x, y) = xy
        assert 0 <= x <= self.width - hotspot.width
        assert 0 <= y <= self.height - hotspot.height

        # TODO: should it check to see whether hotspots overlap each other?
        # Is sensible to _allow_ them to overlap?
        self._hotspots.append((hotspot, xy))

    def remove_hotspot(self, hotspot, xy):
        """
        Remove the hotspot at ``(x, y)``: Any previously rendered image where
        the hotspot was placed is erased from the backing image, and will be
        "undrawn" the next time the virtual device is refreshed. If the
        specified hotspot is not found for ``(x, y)``, a ``ValueError`` is
        raised.
        """
        self._hotspots.remove((hotspot, xy))
        eraser = Image.new(self.mode, hotspot.size)
        self._backing_image.paste(eraser, xy)

    def is_overlapping_viewport(self, hotspot, xy):
        """
        Checks to see if the hotspot at position ``(x, y)``
        is (at least partially) visible according to the
        position of the viewport.
        """
        l1, t1, r1, b1 = calc_bounds(xy, hotspot)
        l2, t2, r2, b2 = calc_bounds(self._position, self._device)
        return range_overlap(l1, r1, l2, r2) and range_overlap(t1, b1, t2, b2)

    def refresh(self):
        should_wait = False
        for hotspot, xy in self._hotspots:
            if hotspot.should_redraw() and self.is_overlapping_viewport(hotspot, xy):
                pool.add_task(hotspot.paste_into, self._backing_image, xy)
                should_wait = True

        if should_wait:
            pool.wait_completion()

        im = self._backing_image.crop(box=self._crop_box())
        if self._dither:
            im = im.convert(self._device.mode)

        self._device.display(im)

    def _crop_box(self):
        (left, top) = self._position
        right = left + self._device.width
        bottom = top + self._device.height

        assert 0 <= left <= right <= self.width
        assert 0 <= top <= bottom <= self.height

        return (left, top, right, bottom)


class hotspot(mixin.capabilities):
    """
    A hotspot (`a place of more than usual interest, activity, or popularity`)
    is a live display which may be added to a virtual viewport - if the hotspot
    and the viewport are overlapping, then the :func:`update` method will be
    automatically invoked when the viewport is being refreshed or its position
    moved (such that an overlap occurs).

    You would either:

        * create a ``hotspot`` instance, suppling a render function (taking an
          :py:mod:`PIL.ImageDraw` object, ``width`` & ``height`` dimensions.
          The render function should draw within a bounding box of ``(0, 0,
          width, height)``, and render a full frame.

        * sub-class ``hotspot`` and override the :func:`should_redraw` and
          :func:`update` methods. This might be more useful for slow-changing
          values where it is not necessary to update every refresh cycle, or
          your implementation is stateful.
    """
    def __init__(self, width, height, draw_fn=None):
        self.capabilities(width, height, rotate=0)  # TODO: set mode?
        self._fn = draw_fn

    def paste_into(self, image, xy):
        im = Image.new(image.mode, self.size)
        draw = ImageDraw.Draw(im)
        self.update(draw)
        image.paste(im, xy)
        del draw

    def should_redraw(self):
        """
        Override this method to return true or false on some condition
        (possibly on last updated member variable) so that for slow changing
        hotspots they are not updated too frequently.
        """
        return True

    def update(self, draw):
        if self._fn:
            self._fn(draw, self.width, self.height)


class snapshot(hotspot):
    """
    A snapshot is a `type of` hotspot, but only updates once in a given
    interval, usually much less frequently than the viewport requests refresh
    updates.
    """
    def __init__(self, width, height, draw_fn=None, interval=1.0):
        assert interval > 0
        super(snapshot, self).__init__(width, height, draw_fn)
        self.interval = interval
        self.last_updated = -interval

    def should_redraw(self):
        """
        Only requests a redraw after ``interval`` seconds have elapsed.
        """
        return perf_counter() - self.last_updated > self.interval

    def paste_into(self, image, xy):
        super(snapshot, self).paste_into(image, xy)
        self.last_updated = perf_counter()


class terminal(object):
    """
    Provides a terminal-like interface to a device (or a device-like object
    that has :class:`mixin.capabilities` characteristics).
    """
    def __init__(self, device, font=None, color="white", bgcolor="black",
                 tabstop=4, line_height=None, animate=True, word_wrap=False):
        self._device = device
        self.font = font or ImageFont.load_default()
        self.default_fgcolor = color
        self.default_bgcolor = bgcolor
        self.animate = animate
        self.tabstop = tabstop
        self.word_wrap = word_wrap

        self._cw, self._ch = (0, 0)
        for i in range(32, 128):
            left, top, w, h = self.font.getbbox(chr(i))
            self._cw = max(w, self._cw)
            self._ch = max(h, self._ch)

        self._ch = line_height or self._ch
        self.width = device.width // self._cw
        self.height = device.height // self._ch
        self.size = (self.width, self.height)
        self.reset()
        self._backing_image = Image.new(self._device.mode, self._device.size,
            self._bgcolor)
        self._canvas = ImageDraw.Draw(self._backing_image)
        self.clear()

        if self.word_wrap:
            self.tw = TextWrapper()
            self.tw.width = self.width
            self.tw.expand_tabs = False
            self.tw.replace_whitespace = False
            self.tw.drop_whitespace = False
            self.tw.break_long_words = True

    def clear(self):
        """
        Clears the display and resets the cursor position to ``(0, 0)``.
        """
        self._cx, self._cy = (0, 0)
        self._canvas.rectangle(self._device.bounding_box,
                               fill=self.default_bgcolor)
        self.flush()

    def println(self, text=""):
        """
        Prints the supplied text to the device, scrolling where necessary.
        The text is always followed by a newline.

        :param text: The text to print.
        :type text: str
        """
        if self.word_wrap:
            # find directives in complete text
            directives = ansi_color.find_directives(text, self)

            # strip ansi from text
            clean_text = ansi_color.strip_ansi_codes(text)

            # wrap clean text
            clean_lines = self.tw.wrap(clean_text)

            # print wrapped text
            index = 0
            for line in clean_lines:
                line_length = len(line)
                y = 0
                while y < line_length:
                    method, args = directives[index]
                    if method == self.putch:
                        y += 1
                    method(*args)
                    index += 1
                self.newline()
        else:
            self.puts(text)
            self.newline()

    def puts(self, text):
        """
        Prints the supplied text, handling special character codes for carriage
        return (\\r), newline (\\n), backspace (\\b) and tab (\\t). ANSI color
        codes are also supported.

        If the ``animate`` flag was set to True (default), then each character
        is flushed to the device, giving the effect of 1970's teletype device.

        :param text: The text to print.
        :type text: str
        """
        for method, args in ansi_color.find_directives(text, self):
            method(*args)

    def putch(self, char):
        """
        Prints the specific character, which must be a valid printable ASCII
        value in the range 32..127 only, or one of carriage return (\\r),
        newline (\\n), backspace (\\b) or tab (\\t).

        :param char: The character to print.
        """
        if char == '\r':
            self.carriage_return()

        elif char == '\n':
            self.newline()

        elif char == '\b':
            self.backspace()

        elif char == '\t':
            self.tab()

        else:
            left, top, w, h = self.font.getbbox(char)
            if self._cx + w >= self._device.width:
                self.newline()

            self.erase()
            self._canvas.text((self._cx, self._cy),
                              text=char,
                              font=self.font,
                              fill=self._fgcolor)

            self._cx += w
            if self.animate:
                self.flush()

    def carriage_return(self):
        """
        Returns the cursor position to the left-hand side without advancing
        downwards.
        """
        self._cx = 0

    def tab(self):
        """
        Advances the cursor position to the next (soft) tabstop.
        """
        soft_tabs = self.tabstop - ((self._cx // self._cw) % self.tabstop)
        for _ in range(soft_tabs):
            self.putch(" ")

    def newline(self):
        """
        Advances the cursor position ot the left hand side, and to the next
        line. If the cursor is on the lowest line, the displayed contents are
        scrolled, causing the top line to be lost.
        """
        self.carriage_return()

        if self._cy + (2 * self._ch) >= self._device.height:
            # Simulate a vertical scroll
            copy = self._backing_image.crop((0, self._ch, self._device.width,
                self._device.height))
            self._backing_image.paste(copy, (0, 0))
            self._canvas.rectangle((0, copy.height, self._device.width,
                self._device.height), fill=self.default_bgcolor)
        else:
            self._cy += self._ch

        self.flush()
        if self.animate:
            sleep(0.2)

    def backspace(self):
        """
        Moves the cursor one place to the left, erasing the character at the
        current position. Cannot move beyond column zero, nor onto the
        previous line.
        """
        if self._cx + self._cw >= 0:
            self.erase()
            self._cx -= self._cw

        self.flush()

    def erase(self):
        """
        Erase the contents of the cursor's current position without moving the
        cursor's position.
        """
        bounds = (self._cx, self._cy, self._cx + self._cw, self._cy + self._ch)
        self._canvas.rectangle(bounds, fill=self._bgcolor)

    def flush(self):
        """
        Cause the current backing store to be rendered on the nominated device.
        """
        self._device.display(self._backing_image)

    def foreground_color(self, value):
        """
        Sets the foreground color.

        :param value: The new color value, either string name or RGB tuple.
        :type value: str or tuple
        """
        self._fgcolor = value

    def background_color(self, value):
        """
        Sets the background color.

        :param value: The new color value, either string name or RGB tuple.
        :type value: str or tuple
        """
        self._bgcolor = value

    def reset(self):
        """
        Resets the foreground and background color value back to the original
        when initialised.
        """
        self._fgcolor = self.default_fgcolor
        self._bgcolor = self.default_bgcolor

    def reverse_colors(self):
        """
        Flips the foreground and background colors.
        """
        self._bgcolor, self._fgcolor = self._fgcolor, self._bgcolor


class history(mixin.capabilities):
    """
    Wraps a device (or emulator) to provide a facility to be able to make a
    savepoint (a point at which the screen display can be "rolled-back" to).

    This is mostly useful for displaying transient error/dialog messages
    which could be subsequently dismissed, reverting back to the previous
    display.
    """
    def __init__(self, device):
        self.capabilities(device.width, device.height, rotate=0,
            mode=device.mode)
        if hasattr(device, "segment_mapper"):
            self.segment_mapper = device.segment_mapper
        self._savepoints = []
        self._device = device
        self._last_image = None

    def display(self, image):
        self._last_image = image.copy()
        self._device.display(image)

    def savepoint(self):
        """
        Copies the last displayed image.
        """
        if self._last_image:
            self._savepoints.append(self._last_image)
            self._last_image = None

    def restore(self, drop=0):
        """
        Restores the last savepoint. If ``drop`` is supplied and greater than
        zero, then that many savepoints are dropped, and the next savepoint is
        restored.

        :param drop:
        :type drop: int
        """
        assert drop >= 0
        while drop > 0:
            self._savepoints.pop()
            drop -= 1

        img = self._savepoints.pop()
        self.display(img)

    def __len__(self):
        """
        Indication of the number of savepoints retained.
        """
        return len(self._savepoints)


class sevensegment(object):
    """
    Abstraction that wraps a device, this class provides a ``text`` property
    which can be used to set and get a text value, which when combined with a
    ``segment_mapper`` sets the correct bit representation for seven-segment
    displays and propagates that onto the underlying device.

    :param device: A device instance.
    :param segment_mapper: An optional function that maps strings into the
        correct representation for the 7-segment physical layout. If not
        provided, the default mapper from compatible devices is used instead.
    :param undefined: The default character to substitute when an unrenderable
        character is supplied to the text property.
    :type undefined: char
    """
    def __init__(self, device, undefined="_", segment_mapper=None):
        self.device = device
        self.undefined = undefined
        self.segment_mapper = segment_mapper or device.segment_mapper
        self._bufsize = device.width * device.height // 8
        self.text = ""

    @property
    def text(self):
        """
        Returns the current state of the text buffer. This may not reflect
        accurately what is displayed on the seven-segment device, as certain
        alpha-numerics and most punctuation cannot be rendered on the limited
        display.
        """
        return self._text_buffer

    @text.setter
    def text(self, value):
        """
        Updates the seven-segment display with the given value. If there is not
        enough space to show the full text, an ``OverflowException`` is raised.

        :param value: The value to render onto the device. Any characters which
            cannot be rendered will be converted into the ``undefined``
            character supplied in the constructor.
        :type value: str
        """
        self._text_buffer = observable(mutable_string(value),
            observer=self._flush)

    def _flush(self, buf):
        data = bytearray(self.segment_mapper(buf, notfound=self.undefined)
            ).ljust(self._bufsize, b'\0')

        if len(data) > self._bufsize:
            raise OverflowError(
                f"Device's capabilities insufficient for value '{buf}'")

        with canvas(self.device) as draw:
            for x, byte in enumerate(reversed(data)):
                for y in range(8):
                    if byte & 0x01:
                        draw.point((x, y), fill="white")
                    byte >>= 1


class character(object):
    """
    Abstraction that wraps a device, this class provides a ``text`` property
    which can be used to set and get a text value allowing the device to be
    treated as a character style display such as the HD44780 LCD

    If the device is actually a character style device, be careful to provide
    a font that adheres to the pixel dimensions of the display.

    :param device: A device instance.
    :param font: The font to be used to paint the characters within the ``text``
        property.  If the device contains a font (e.g. hd44780, ws0010) it will
        be used as the default if no font is provided.
    :type font: `PIL.ImageFont` object
    :param undefined: The default character to substitute when an unrenderable
        character is supplied to the text property.
    :type undefined: char

    .. versionadded:: 1.15.0
    """

    def __init__(self, device, font=None, undefined="_"):
        self.device = device
        self._undefined = undefined

        self.font = font if font else device.font if hasattr(device, 'font') else None
        assert self.font, 'No font available'
        self.text = ''

    @property
    def text(self):
        """
        Returns the current state of the text buffer. This may not reflect
        accurately what is displayed on the device if the font does
        not have a symbol for a requested text value.
        """
        return self._text_buffer

    @text.setter
    def text(self, value):
        """
        Updates the display with the given value.

        :param value: The value to render onto the device. Any characters which
            cannot be rendered will be converted into the ``undefined``
            character supplied in the constructor. Newline characters '\n' work
            as expected but no other control characters (e.g. \r) are honored.
        :type value: str
        """
        self._text_buffer = observable(mutable_string(value),
            observer=self._flush)

    def _flush(self, buf):
        # Replace any characters that are not in the font with the undefined character
        buf = ''.join([i if i == '\n' or self.font.getlength(i) > 0 else self._undefined for i in buf])

        # Draw text onto display's image using the provided font
        with canvas(self.device) as draw:
            # Place text
            draw.text((0, 0), buf, fill='white', font=self.font, spacing=0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-20 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:mod:`luma.core.util` module.
"""


from luma.core import util

from helpers import get_reference_file


test_config_file = get_reference_file('config-test.txt')


def test_mutablestring():
    f = util.mutable_string('bar')
    f[1] = '2'
    assert f == 'b2r'
    assert repr(f) == repr('b2r')
    assert len(f) == 3
    assert hash(f) is not None


def test_mutablestring_unicode():
    f = util.mutable_string(u'bazül')
    f[4] = 'L'
    assert f == u'baz\xfcL'
    assert repr(f) == repr(u'baz\xfcL')
    assert len(f) == 5


def test_from_16_to_8_to_16():
    """
    Test the conversion from 16 bit to 8 bit values and back.  Ensure that conversion
    from 8 to 16 bit takes into account whether the 16 bit value is negative or
    positive
    """
    data = [1, 2, 3, 2**15, 2**16 - 1]
    expect = [0, 1, 0, 2, 0, 3, 0x80, 0, 0xFF, 0xFF]

    result = util.from_16_to_8(data)
    assert result == expect

    expect = [1, 2, 3, -32768, -1]
    result = util.from_8_to_16(result)
    assert result == expect


def test_bytes_to_nibbles():
    """
    Test the conversion from 8 bit values into 4 bit (nibbles)
    """
    data = [0, 1, 2, 3, 0xFF, 0x7F]
    expect = [0, 0, 0, 1, 0, 2, 0, 3, 15, 15, 7, 15]

    result = util.bytes_to_nibbles(data)
    assert result == expect



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.


from luma.core.lib import spidev, rpi_gpio


@rpi_gpio
class RpiGpioTest(object):
    pass


@spidev
class SpiDevTest(object):
    pass


@spidev
@rpi_gpio
class MultiLibTest(object):
    pass


def assertMethod(obj, method):
    assert hasattr(obj, method)
    assert callable(getattr(obj, method))


def test_rpio_gpio():
    t = RpiGpioTest()
    assertMethod(t, '__rpi_gpio__')


def test_spidev():
    t = SpiDevTest()
    assertMethod(t, '__spidev__')


def test_multi():
    t = MultiLibTest()
    for method in ['__spidev__', '__rpi_gpio__']:
        assertMethod(t, method)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019-2020 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.interface.serial.ftdi_i2c` class.
"""

import pytest
from unittest.mock import Mock, patch
from luma.core.interface.serial import ftdi_i2c
from helpers import fib
import luma.core.error


@patch('pyftdi.i2c.I2cController')
def test_init(mock_controller):
    instance = Mock()
    instance.get_port = Mock()
    mock_controller.side_effect = [instance]

    ftdi_i2c(device='ftdi://dummy', address='0xFF')
    mock_controller.assert_called_with()
    instance.configure.assert_called_with('ftdi://dummy')
    instance.get_port.assert_called_with(0xFF)


@patch('pyftdi.i2c.I2cController')
def test_command(mock_controller):
    cmds = [3, 1, 4, 2]
    port = Mock()
    instance = Mock()
    instance.get_port = Mock(return_value=port)
    mock_controller.side_effect = [instance]

    serial = ftdi_i2c(device='ftdi://dummy', address=0x3C)
    serial.command(*cmds)
    port.write_to.assert_called_once_with(0x00, cmds)


@patch('pyftdi.i2c.I2cController')
def test_data(mock_controller):
    data = list(fib(100))
    port = Mock()
    instance = Mock()
    instance.get_port = Mock(return_value=port)
    mock_controller.side_effect = [instance]

    serial = ftdi_i2c(device='ftdi://dummy', address=0x3C)
    serial.data(data)
    port.write_to.assert_called_once_with(0x40, data)


@patch('pyftdi.i2c.I2cController')
def test_cleanup(mock_controller):
    port = Mock()
    instance = Mock()
    instance.get_port = Mock(return_value=port)
    mock_controller.side_effect = [instance]

    serial = ftdi_i2c(device='ftdi://dummy', address=0x3C)
    serial.cleanup()
    instance.terminate.assert_called_once_with()


@patch('pyftdi.i2c.I2cController')
def test_init_device_address_error(mock_controller):
    address = 'foo'
    with pytest.raises(luma.core.error.DeviceAddressError) as ex:
        ftdi_i2c(device='ftdi://dummy', address=address)
    assert str(ex.value) == f'I2C device address invalid: {address}'
